#ifndef SQUARE_Q2_TRANSFORMATION_HPP
#define SQUARE_Q2_TRANSFORMATION_HPP

#include <algebra/TinyMatrix.hpp>
#include <algebra/TinyVector.hpp>

template <size_t GivenDimension>
class SquareQ2Transformation;

template <>
class SquareQ2Transformation<2>
{
 public:
  constexpr static size_t Dimension      = 2;
  constexpr static size_t NumberOfPoints = 9;

 private:
  // Points are defined as follows (image of the unit square)
  //
  //  a02--a12--a22
  //   |         |
  //  a01  a11  a21
  //   |         |
  //  a00--a10--a20
  TinyMatrix<Dimension, NumberOfPoints> m_points_images;

  PUGS_INLINE
  double
  _w0(const double& x) const
  {
    return 0.5 * (x - 1) * x;
  }

  PUGS_INLINE
  double
  _w1(const double& x) const
  {
    return (1 - x) * (1 + x);
  }

  PUGS_INLINE
  double
  _w2(const double& x) const
  {
    return 0.5 * x * (1 + x);
  }

  PUGS_INLINE
  double
  _dw0(const double& x) const
  {
    return x - 0.5;
  }

  PUGS_INLINE
  double
  _dw1(const double& x) const
  {
    return -2 * x;
  }

  PUGS_INLINE
  double
  _dw2(const double& x) const
  {
    return x + 0.5;
  }

 public:
  PUGS_INLINE
  TinyVector<Dimension>
  operator()(const TinyVector<2>& X_hat) const
  {
    const double& x_hat = X_hat[0];
    const double& y_hat = X_hat[1];

    const double w0x = _w0(x_hat);
    const double w1x = _w1(x_hat);
    const double w2x = _w2(x_hat);

    const double w0y = _w0(y_hat);
    const double w1y = _w1(y_hat);
    const double w2y = _w2(y_hat);

    TinyVector<NumberOfPoints> Wij{w0x * w0y, w0x * w1y, w0x * w2y,   //
                                   w1x * w0y, w1x * w1y, w1x * w2y,   //
                                   w2x * w0y, w2x * w1y, w2x * w2y};

    return m_points_images * Wij;
  }

  PUGS_INLINE double
  jacobianDeterminant(const TinyVector<Dimension>& X_hat) const
  {
    const double& x_hat = X_hat[0];
    const double& y_hat = X_hat[1];

    const double w0x = _w0(x_hat);
    const double w1x = _w1(x_hat);
    const double w2x = _w2(x_hat);

    const double w0y = _w0(y_hat);
    const double w1y = _w1(y_hat);
    const double w2y = _w2(y_hat);

    const double dx_w0x = _dw0(x_hat);
    const double dx_w1x = _dw1(x_hat);
    const double dx_w2x = _dw2(x_hat);

    const double dy_w0y = _dw0(y_hat);
    const double dy_w1y = _dw1(y_hat);
    const double dy_w2y = _dw2(y_hat);

    TinyVector<NumberOfPoints> dx_Wij{dx_w0x * w0y, dx_w0x * w1y, dx_w0x * w2y,   //
                                      dx_w1x * w0y, dx_w1x * w1y, dx_w1x * w2y,   //
                                      dx_w2x * w0y, dx_w2x * w1y, dx_w2x * w2y};

    TinyVector<NumberOfPoints> dy_Wij{w0x * dy_w0y, w0x * dy_w1y, w0x * dy_w2y,   //
                                      w1x * dy_w0y, w1x * dy_w1y, w1x * dy_w2y,   //
                                      w2x * dy_w0y, w2x * dy_w1y, w2x * dy_w2y};

    const TinyMatrix<Dimension, Dimension> J{m_points_images * dx_Wij,   //
                                             m_points_images * dy_Wij};
    return det(J);
  }

  PUGS_INLINE
  SquareQ2Transformation(const TinyVector<Dimension>& a00,
                         const TinyVector<Dimension>& a01,
                         const TinyVector<Dimension>& a02,
                         const TinyVector<Dimension>& a10,
                         const TinyVector<Dimension>& a11,
                         const TinyVector<Dimension>& a12,
                         const TinyVector<Dimension>& a20,
                         const TinyVector<Dimension>& a21,
                         const TinyVector<Dimension>& a22)
    : m_points_images{a00, a01, a02, a10, a11, a12, a20, a21, a22}
  {}

  ~SquareQ2Transformation() = default;
};

#endif   // SQUARE_Q2_TRANSFORMATION_HPP
