#ifndef SQUARE_Q3_TRANSFORMATION_HPP
#define SQUARE_Q3_TRANSFORMATION_HPP

#include <algebra/TinyMatrix.hpp>
#include <algebra/TinyVector.hpp>

template <size_t GivenDimension>
class SquareQ3Transformation;

template <>
class SquareQ3Transformation<2>
{
 public:
  constexpr static size_t Dimension      = 2;
  constexpr static size_t NumberOfPoints = 16;

 private:
  // Points are defined as follows (image of the unit square)
  //  a03--a13--a23--a33
  //   |    |    |    |
  //  a02--a12--a22--a32
  //   |    |    |    |
  //  a01--a11--a21--a31
  //   |    |    |    |
  //  a00--a10--a20--a30
  TinyMatrix<Dimension, NumberOfPoints> m_points_images;

  PUGS_INLINE
  double
  _w0(const double& x) const
  {
    return -9. / 16 * (x + 1. / 3) * (x - 1. / 3) * (x - 1);
  }

  PUGS_INLINE
  double
  _w1(const double& x) const
  {
    return 27. / 16 * (x + 1) * (x - 1. / 3) * (x - 1);
  }

  PUGS_INLINE
  double
  _w2(const double& x) const
  {
    return -27. / 16 * (x + 1) * (x + 1. / 3) * (x - 1);
  }

  PUGS_INLINE
  double
  _w3(const double& x) const
  {
    return 9. / 16 * (x + 1) * (x + 1. / 3) * (x - 1. / 3);
  }

  PUGS_INLINE
  double
  _dw0(const double& x) const
  {
    return -9. / 16 * (3 * x * x - 2 * x - 1. / 9);
  }

  PUGS_INLINE
  double
  _dw1(const double& x) const
  {
    return 27. / 16 * (3 * x * x - 2. / 3 * x - 1);
  }

  PUGS_INLINE
  double
  _dw2(const double& x) const
  {
    return -27. / 16 * (3 * x * x + 2. / 3 * x - 1);
  }

  PUGS_INLINE
  double
  _dw3(const double& x) const
  {
    return 9. / 16 * (3 * x * x + 2 * x - 1. / 9);
  }

 public:
  PUGS_INLINE
  TinyVector<Dimension>
  operator()(const TinyVector<2>& X_hat) const
  {
    const double& x_hat = X_hat[0];
    const double& y_hat = X_hat[1];

    const double w0x = _w0(x_hat);
    const double w1x = _w1(x_hat);
    const double w2x = _w2(x_hat);
    const double w3x = _w3(x_hat);

    const double w0y = _w0(y_hat);
    const double w1y = _w1(y_hat);
    const double w2y = _w2(y_hat);
    const double w3y = _w3(y_hat);

    TinyVector<NumberOfPoints> Wij{w0x * w0y, w0x * w1y, w0x * w2y, w0x * w3y,   //
                                   w1x * w0y, w1x * w1y, w1x * w2y, w1x * w3y,   //
                                   w2x * w0y, w2x * w1y, w2x * w2y, w2x * w3y,   //
                                   w3x * w0y, w3x * w1y, w3x * w2y, w3x * w3y};

    //  std::clog << "W=" << Wij << '\n';
    // std::clog << "points=" << m_points_images << '\n';

    return m_points_images * Wij;
  }

  PUGS_INLINE double
  jacobianDeterminant(const TinyVector<Dimension>& X_hat) const
  {
    const double& x_hat = X_hat[0];
    const double& y_hat = X_hat[1];

    const double w0x = _w0(x_hat);
    const double w1x = _w1(x_hat);
    const double w2x = _w2(x_hat);
    const double w3x = _w3(x_hat);

    const double w0y = _w0(y_hat);
    const double w1y = _w1(y_hat);
    const double w2y = _w2(y_hat);
    const double w3y = _w3(y_hat);

    const double dx_w0x = _dw0(x_hat);
    const double dx_w1x = _dw1(x_hat);
    const double dx_w2x = _dw2(x_hat);
    const double dx_w3x = _dw3(x_hat);

    const double dy_w0y = _dw0(y_hat);
    const double dy_w1y = _dw1(y_hat);
    const double dy_w2y = _dw2(y_hat);
    const double dy_w3y = _dw3(y_hat);

    TinyVector<NumberOfPoints> dx_Wij{dx_w0x * w0y, dx_w0x * w1y, dx_w0x * w2y, dx_w0x * w3y,   //
                                      dx_w1x * w0y, dx_w1x * w1y, dx_w1x * w2y, dx_w1x * w3y,   //
                                      dx_w2x * w0y, dx_w2x * w1y, dx_w2x * w2y, dx_w2x * w3y,   //
                                      dx_w3x * w0y, dx_w3x * w1y, dx_w3x * w2y, dx_w3x * w3y};

    TinyVector<NumberOfPoints> dy_Wij{w0x * dy_w0y, w0x * dy_w1y, w0x * dy_w2y, w0x * dy_w3y,   //
                                      w1x * dy_w0y, w1x * dy_w1y, w1x * dy_w2y, w1x * dy_w3y,   //
                                      w2x * dy_w0y, w2x * dy_w1y, w2x * dy_w2y, w2x * dy_w3y,   //
                                      w3x * dy_w0y, w3x * dy_w1y, w3x * dy_w2y, w3x * dy_w3y};

    const TinyMatrix<Dimension, Dimension> J{m_points_images * dx_Wij,   //
                                             m_points_images * dy_Wij};
    return det(J);
  }

  PUGS_INLINE
  SquareQ3Transformation(const TinyVector<Dimension>& a00,
                         const TinyVector<Dimension>& a01,
                         const TinyVector<Dimension>& a02,
                         const TinyVector<Dimension>& a03,
                         const TinyVector<Dimension>& a10,
                         const TinyVector<Dimension>& a11,
                         const TinyVector<Dimension>& a12,
                         const TinyVector<Dimension>& a13,
                         const TinyVector<Dimension>& a20,
                         const TinyVector<Dimension>& a21,
                         const TinyVector<Dimension>& a22,
                         const TinyVector<Dimension>& a23,
                         const TinyVector<Dimension>& a30,
                         const TinyVector<Dimension>& a31,
                         const TinyVector<Dimension>& a32,
                         const TinyVector<Dimension>& a33)
    : m_points_images{a00, a01, a02, a03, a10, a11, a12, a13, a20, a21, a22, a23, a30, a31, a32, a33}
  {}

  ~SquareQ3Transformation() = default;
};

#endif   // SQUARE_Q3_TRANSFORMATION_HPP
