#ifndef TRIANGLE_P3_TRANSFORMATION_HPP
#define TRIANGLE_P3_TRANSFORMATION_HPP

#include <algebra/TinyMatrix.hpp>
#include <algebra/TinyVector.hpp>

template <size_t GivenDimension>
class TriangleP3Transformation;

template <>
class TriangleP3Transformation<2>
{
 public:
  constexpr static size_t Dimension      = 2;
  constexpr static size_t NumberOfPoints = 10;

 private:
  // Points are defined as follows (image of the unit triangle with a right angle in a0)
  //
  //   a2
  //   |  `
  //  a220  a122
  //   |     `
  //  a200  a012    a112
  //   |         `
  //   a0- a001- a001 -a1

  TinyMatrix<Dimension, NumberOfPoints> m_points_images;

  PUGS_INLINE
  double
  _lambda0(const TinyVector<Dimension>& X_hat) const
  {
    const double& x_hat = X_hat[0];
    const double& y_hat = X_hat[1];

    return 1 - x_hat - y_hat;
  }

  PUGS_INLINE
  double
  _lambda1(const TinyVector<Dimension>& X_hat) const
  {
    const double& x_hat = X_hat[0];

    return x_hat;
  }

  PUGS_INLINE
  double
  _lambda2(const TinyVector<Dimension>& X_hat) const
  {
    const double& y_hat = X_hat[1];

    return y_hat;
  }

  PUGS_INLINE
  double
  _dx_lambda0(const TinyVector<Dimension>&) const
  {
    return -1;
  }

  PUGS_INLINE
  double
  _dy_lambda0(const TinyVector<Dimension>&) const
  {
    return -1;
  }

  PUGS_INLINE
  double
  _dx_lambda1(const TinyVector<Dimension>&) const
  {
    return 1;
  }

  PUGS_INLINE
  double
  _dy_lambda1(const TinyVector<Dimension>&) const
  {
    return 0;
  }

  PUGS_INLINE
  double
  _dx_lambda2(const TinyVector<Dimension>&) const
  {
    return 0;
  }

  PUGS_INLINE
  double
  _dy_lambda2(const TinyVector<Dimension>&) const
  {
    return 1;
  }

  // Nodal basis functions have the form lambda*(2*lambda -1)
  PUGS_INLINE
  double
  _wa0(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);

    return 1 / 2. * lambda0 * (-1 + 3 * lambda0) * (-2 + 3 * lambda0);
  }

  PUGS_INLINE
  double
  _wa1(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda1 = _lambda1(X_hat);
    return 1 / 2. * lambda1 * (-1 + 3 * lambda1) * (-2 + 3 * lambda1);
  }

  PUGS_INLINE
  double
  _wa2(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2 = _lambda2(X_hat);

    return 1 / 2. * lambda2 * (-1 + 3 * lambda2) * (-2 + 3 * lambda2);
  }

  // Edge basis functions have the form ???
  PUGS_INLINE
  double
  _wa001(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);
    const double lambda1 = _lambda1(X_hat);

    return 27. / 2 * lambda0 * lambda1 * (lambda0 - 1. / 3);
  }
  PUGS_INLINE
  double
  _wa011(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);
    const double lambda1 = _lambda1(X_hat);

    return 27. / 2 * lambda0 * lambda1 * (lambda1 - 1. / 3);
  }

  PUGS_INLINE
  double
  _wa112(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda1 = _lambda1(X_hat);
    const double lambda2 = _lambda2(X_hat);
    return 27. / 2 * lambda1 * lambda2 * (lambda1 - 1. / 3);
  }

  PUGS_INLINE
  double
  _wa122(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda1 = _lambda1(X_hat);
    const double lambda2 = _lambda2(X_hat);
    return 27. / 2 * lambda1 * lambda2 * (lambda2 - 1. / 3);
  }

  PUGS_INLINE
  double
  _wa220(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);
    const double lambda2 = _lambda2(X_hat);
    return 27. / 2 * lambda0 * lambda2 * (lambda2 - 1. / 3);
  }

  PUGS_INLINE
  double
  _wa200(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);
    const double lambda2 = _lambda2(X_hat);
    return 27. / 2 * lambda0 * lambda2 * (lambda0 - 1. / 3);
  }

  PUGS_INLINE
  double
  _wa012(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0 = _lambda0(X_hat);
    const double lambda1 = _lambda1(X_hat);
    const double lambda2 = _lambda2(X_hat);

    return 27. * lambda0 * lambda1 * lambda2;
  }
  /////////////////////
  PUGS_INLINE
  double
  _dx_wa0(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda0) * (-2 + 3 * lambda0) + 3 * lambda0 * (-2 + 3 * lambda0) +
            3 * lambda0 * (-1 + 3 * lambda0)) *
           dx_lambda0;
  }

  PUGS_INLINE
  double
  _dy_wa0(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda0) * (-2 + 3 * lambda0) + 3 * lambda0 * (-2 + 3 * lambda0) +
            3 * lambda0 * (-1 + 3 * lambda0)) *
           dy_lambda0;
  }

  PUGS_INLINE
  double
  _dx_wa1(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda1) * (-2 + 3 * lambda1) + 3 * lambda1 * (-2 + 3 * lambda1) +
            3 * lambda1 * (-1 + 3 * lambda1)) *
           dx_lambda1;
  }
  PUGS_INLINE
  double
  _dy_wa1(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda1) * (-2 + 3 * lambda1) + 3 * lambda1 * (-2 + 3 * lambda1) +
            3 * lambda1 * (-1 + 3 * lambda1)) *
           dy_lambda1;
  }
  PUGS_INLINE
  double
  _dx_wa2(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda2) * (-2 + 3 * lambda2) + 3 * lambda2 * (-2 + 3 * lambda2) +
            3 * lambda2 * (-1 + 3 * lambda2)) *
           dx_lambda2;
  }
  PUGS_INLINE
  double
  _dy_wa2(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);
    return 1. / 2 *
           ((-1 + 3 * lambda2) * (-2 + 3 * lambda2) + 3 * lambda2 * (-2 + 3 * lambda2) +
            3 * lambda2 * (-1 + 3 * lambda2)) *
           dy_lambda2;
  }
  ///////////////////
  PUGS_INLINE
  double
  _dx_wa001(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda0) * lambda0 * dx_lambda1 + lambda1 * (2 * lambda0 - 1. / 3) * dx_lambda0);
  }
  PUGS_INLINE
  double
  _dy_wa001(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda0) * lambda0 * dy_lambda1 + lambda1 * (2 * lambda0 - 1. / 3) * dy_lambda0);
  }

  PUGS_INLINE
  double
  _dx_wa011(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda1) * lambda1 * dx_lambda0 + lambda0 * (2 * lambda1 - 1. / 3) * dx_lambda1);
  }

  PUGS_INLINE
  double
  _dy_wa011(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda1) * lambda1 * dy_lambda0 + lambda0 * (2 * lambda1 - 1. / 3) * dy_lambda1);
  }

  PUGS_INLINE
  double
  _dx_wa122(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda2) * lambda2 * dx_lambda1 + lambda1 * (2 * lambda2 - 1. / 3) * dx_lambda2);
  }

  PUGS_INLINE
  double
  _dy_wa122(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda2) * lambda2 * dy_lambda1 + lambda1 * (2 * lambda2 - 1. / 3) * dy_lambda2);
  }

  PUGS_INLINE
  double
  _dx_wa112(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda1) * lambda1 * dx_lambda2 + lambda2 * (2 * lambda1 - 1. / 3) * dx_lambda1);
  }

  PUGS_INLINE
  double
  _dy_wa112(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda1) * lambda1 * dy_lambda2 + lambda2 * (2 * lambda1 - 1. / 3) * dy_lambda1);
  }

  PUGS_INLINE
  double
  _dx_wa220(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda2) * lambda2 * dx_lambda0 + lambda0 * (2 * lambda2 - 1. / 3) * dx_lambda2);
  }

  PUGS_INLINE
  double
  _dy_wa220(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda2) * lambda2 * dy_lambda0 + lambda0 * (2 * lambda2 - 1. / 3) * dy_lambda2);
  }

  PUGS_INLINE
  double
  _dx_wa200(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda0) * lambda0 * dx_lambda2 + lambda2 * (2 * lambda0 - 1. / 3) * dx_lambda0);
  }
  PUGS_INLINE
  double
  _dy_wa200(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);

    return 27. / 2 * ((-1. / 3 + lambda0) * lambda0 * dy_lambda2 + lambda2 * (2 * lambda0 - 1. / 3) * dy_lambda0);
  }

  PUGS_INLINE
  double
  _dx_wa012(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dx_lambda0 = _dx_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dx_lambda1 = _dx_lambda1(X_hat);
    const double lambda2    = _lambda2(X_hat);
    const double dx_lambda2 = _dx_lambda2(X_hat);

    return 27 * (dx_lambda0 * lambda1 * lambda2 + lambda0 * dx_lambda1 * lambda2 + lambda0 * lambda1 * dx_lambda2);
  }

  PUGS_INLINE
  double
  _dy_wa012(const TinyVector<Dimension>& X_hat) const
  {
    const double lambda0    = _lambda0(X_hat);
    const double dy_lambda0 = _dy_lambda0(X_hat);
    const double lambda1    = _lambda1(X_hat);
    const double dy_lambda1 = _dy_lambda1(X_hat);
    const double lambda2    = _lambda2(X_hat);
    const double dy_lambda2 = _dy_lambda2(X_hat);

    return 27 * (dy_lambda0 * lambda1 * lambda2 + lambda0 * dy_lambda1 * lambda2 + lambda0 * lambda1 * dy_lambda2);
  }

 public:
  PUGS_INLINE
  TinyVector<Dimension>
  operator()(const TinyVector<2>& X_hat) const
  {
    TinyVector<NumberOfPoints> Wij{_wa0(X_hat),   _wa1(X_hat),   _wa2(X_hat),     //
                                   _wa112(X_hat), _wa122(X_hat), _wa220(X_hat),   //
                                   _wa200(X_hat), _wa001(X_hat), _wa011(X_hat),   //
                                   _wa012(X_hat)};

    return m_points_images * Wij;
  }

  double
  jacobianDeterminant(const TinyVector<2>& X_hat) const
  {
    const TinyVector<NumberOfPoints> dx_Wij{_dx_wa0(X_hat),   _dx_wa1(X_hat),   _dx_wa2(X_hat),     //
                                            _dx_wa112(X_hat), _dx_wa122(X_hat), _dx_wa220(X_hat),   //
                                            _dx_wa200(X_hat), _dx_wa001(X_hat), _dx_wa011(X_hat),   //
                                            _dx_wa012(X_hat)};

    const TinyVector<NumberOfPoints> dy_Wij{_dy_wa0(X_hat),   _dy_wa1(X_hat),   _dy_wa2(X_hat),     //
                                            _dy_wa112(X_hat), _dy_wa122(X_hat), _dy_wa220(X_hat),   //
                                            _dy_wa200(X_hat), _dy_wa001(X_hat), _dy_wa011(X_hat),   //
                                            _dy_wa012(X_hat)};

    const TinyMatrix<Dimension, Dimension> J{m_points_images * dx_Wij,   //
                                             m_points_images * dy_Wij};
    return det(J);
  }

  PUGS_INLINE
  TriangleP3Transformation(const TinyVector<Dimension>& a0,
                           const TinyVector<Dimension>& a1,
                           const TinyVector<Dimension>& a2,
                           const TinyVector<Dimension>& a112,
                           const TinyVector<Dimension>& a122,
                           const TinyVector<Dimension>& a220,
                           const TinyVector<Dimension>& a200,
                           const TinyVector<Dimension>& a001,
                           const TinyVector<Dimension>& a011,
                           const TinyVector<Dimension>& a012)

    : m_points_images{a0, a1, a2, a112, a122, a220, a200, a001, a011, a012}
  {}

  ~TriangleP3Transformation() = default;
};

#endif   // TRIANGLE_P3_TRANSFORMATION_HPP
