#include <language/modules/VariationalSchemeModule.hpp>

#include <language/modules/MeshModuleTypes.hpp>
#include <language/modules/ModuleRepository.hpp>
#include <language/modules/SchemeModuleTypes.hpp>
#include <language/utils/BuiltinFunctionEmbedder.hpp>
#include <language/utils/TypeDescriptor.hpp>
#include <scheme/DiscreteFunctionUtils.hpp>
#include <scheme/P1P0AnalyticVariationalSolver.hpp>
#include <scheme/P1P0VariationalSolver.hpp>
#include <scheme/VariationalSolver.hpp>
#include <scheme/VariationalSolverDevelopedReconstruction.hpp>
#include <scheme/VariationalSolverO1.hpp>
#include <scheme/WallBoundaryConditionDescriptor.hpp>

#include <memory>

VariationalSchemeModule::VariationalSchemeModule()
{
  this->_addBuiltinFunction("variational_solver",
                            std::function(

                              [](const size_t& degree, const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverHandler{getCommonMesh({rho, u, E, c, a})}
                                  .solver()
                                  .apply(degree, dt, VariationalSolverHandler::VelocityBCTreatment::elimination, rho, u,
                                         E, c, a, bc_descriptor_list, {});
                              }

                              ));

  this->_addBuiltinFunction("variational_solver_with_source",
                            std::function(

                              [](const size_t& degree, const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt, const FunctionSymbolId& function_id)
                                -> std::tuple<std::shared_ptr<const MeshVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverHandler{getCommonMesh({rho, u, E, c, a})}
                                  .solver()
                                  .apply(degree, dt, VariationalSolverHandler::VelocityBCTreatment::elimination, rho, u,
                                         E, c, a, bc_descriptor_list, function_id);
                              }

                              ));

  this->_addBuiltinFunction("variational_solver_dr",
                            std::function(

                              [](const size_t& degree, const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverDevelopedReconstructionHandler{getCommonMesh({rho, u, E, c, a})}
                                  .solver()
                                  .apply(degree, dt,
                                         VariationalSolverDevelopedReconstructionHandler::VelocityBCTreatment::
                                           elimination,
                                         rho, u, E, c, a, bc_descriptor_list, {});
                              }

                              ));

  this->_addBuiltinFunction("variational_solver_dr_with_source",
                            std::function(

                              [](const size_t& degree, const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt, const FunctionSymbolId& function_id)
                                -> std::tuple<std::shared_ptr<const MeshVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>,
                                              std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverDevelopedReconstructionHandler{getCommonMesh({rho, u, E, c, a})}
                                  .solver()
                                  .apply(degree, dt,
                                         VariationalSolverDevelopedReconstructionHandler::VelocityBCTreatment::
                                           elimination,
                                         rho, u, E, c, a, bc_descriptor_list, function_id);
                              }

                              ));

  this->_addBuiltinFunction("variational_solver_o1",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverO1Handler{getCommonMesh({rho, u, E, c, a, p})}
                                  .solver()
                                  .apply(dt, VariationalSolverO1Handler::VelocityBCTreatment::elimination, rho, u, E, c,
                                         a, p, bc_descriptor_list);
                              }

                              ));

  this->_addBuiltinFunction("penalized_variational_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return VariationalSolverO1Handler{getCommonMesh({rho, u, E, c, a, p})}
                                  .solver()
                                  .apply(dt, VariationalSolverO1Handler::VelocityBCTreatment::penalty, rho, u, E, c, a,
                                         p, bc_descriptor_list);
                              }

                              ));

  this->_addBuiltinFunction("p1_p0_analytic_variational_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return P1P0AnalyticVariationalSolverHandler{getCommonMesh({rho, u, E, c, a, p})}
                                  .solver()
                                  .apply(dt, P1P0AnalyticVariationalSolverHandler::VelocityBCTreatment::elimination,
                                         rho, u, E, c, a, p, bc_descriptor_list);
                              }

                              ));

  this->_addBuiltinFunction("p1_p0_variational_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return P1P0VariationalSolverHandler{getCommonMesh({rho, u, E, c, a, p})}
                                  .solver()
                                  .apply(dt, P1P0VariationalSolverHandler::VelocityBCTreatment::elimination, rho, u, E,
                                         c, a, p, bc_descriptor_list);
                              }

                              ));

  this->_addBuiltinFunction("p1_p0_analytic_penalized_variational_solver",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& rho,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& u,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& E,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& c,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& a,
                                 const std::shared_ptr<const DiscreteFunctionVariant>& p,
                                 const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>&
                                   bc_descriptor_list,
                                 const double& dt) -> std::tuple<std::shared_ptr<const MeshVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>,
                                                                 std::shared_ptr<const DiscreteFunctionVariant>> {
                                return P1P0AnalyticVariationalSolverHandler{getCommonMesh({rho, u, E, c, a, p})}
                                  .solver()
                                  .apply(dt, P1P0AnalyticVariationalSolverHandler::VelocityBCTreatment::penalty, rho, u,
                                         E, c, a, p, bc_descriptor_list);
                              }

                              ));

  this->_addBuiltinFunction("variational_acoustic_dt",
                            std::function(

                              [](const std::shared_ptr<const DiscreteFunctionVariant>& c) -> double {
                                return variational_acoustic_dt(c);
                              }

                              ));
}

void
VariationalSchemeModule::registerOperators() const
{}

void
VariationalSchemeModule::registerCheckpointResume() const
{}

ModuleRepository::Subscribe<VariationalSchemeModule> variational_scheme_module;
