#ifndef INTEGRATE_ON_CELLS_HPP
#define INTEGRATE_ON_CELLS_HPP

#include <analysis/IQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/CubeTransformation.hpp>
#include <geometry/LineTransformation.hpp>
#include <geometry/PrismTransformation.hpp>
#include <geometry/PyramidTransformation.hpp>
#include <geometry/SquareQ2Transformation.hpp>
#include <geometry/SquareQ3Transformation.hpp>
#include <geometry/SquareTransformation.hpp>
#include <geometry/TetrahedronTransformation.hpp>
#include <geometry/TriangleP2Transformation.hpp>
#include <geometry/TriangleP3Transformation.hpp>
#include <geometry/TriangleTransformation.hpp>
#include <language/utils/PugsFunctionAdapter.hpp>
#include <mesh/CellType.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ItemArray.hpp>
#include <mesh/ItemId.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>
#include <mesh/MeshTraits.hpp>
#include <utils/Array.hpp>

class FunctionSymbolId;

template <typename T>
class IntegrateOnCells;
template <typename OutputType, typename InputType>
class IntegrateOnCells<OutputType(InputType)> : public PugsFunctionAdapter<OutputType(InputType)>
{
 private:
  using Adapter = PugsFunctionAdapter<OutputType(InputType)>;
  using Rd      = InputType;

  template <size_t Dimension>
  class AllCells
  {
   private:
    const Connectivity<Dimension>& m_connectivity;

   public:
    using index_type = CellId;

    PUGS_INLINE
    CellId
    cellId(const CellId cell_id) const
    {
      return cell_id;
    }

    PUGS_INLINE
    size_t
    size() const
    {
      return m_connectivity.numberOfCells();
    }

    PUGS_INLINE
    AllCells(const Connectivity<Dimension>& connectivity) : m_connectivity{connectivity} {}
  };

  class CellList
  {
   private:
    const Array<const CellId>& m_cell_list;

   public:
    using index_type = Array<const CellId>::index_type;

    PUGS_INLINE
    CellId
    cellId(const index_type index) const
    {
      return m_cell_list[index];
    }

    PUGS_INLINE
    size_t
    size() const
    {
      return m_cell_list.size();
    }

    PUGS_INLINE
    CellList(const Array<const CellId>& cell_list) : m_cell_list{cell_list} {}
  };

  template <MeshConcept MeshType, typename OutputArrayT, typename ListTypeT>
  static PUGS_INLINE void
  _tensorialIntegrateTo(const FunctionSymbolId& function_symbol_id,
                        const IQuadratureDescriptor& quadrature_descriptor,
                        const MeshType& mesh,
                        const ListTypeT& cell_list,
                        OutputArrayT& value)
  {
    static_assert((is_polygonal_mesh_v<MeshType> or is_polynomial_mesh_v<MeshType>)   //
                  and(MeshType::Dimension == InputType::Dimension));

    static_assert(is_polygonal_mesh_v<MeshType> or (is_polynomial_mesh_v<MeshType> and (MeshType::Dimension == 2)),
                  "Polynomial meshes are only supported in dimension 2");

    if constexpr (is_polynomial_mesh_v<MeshType>) {
      if (mesh.degree() > 3) {
        throw NotImplementedError("IntegrateOnCells only support polynomial meshes of degree <= 3");
      }
    }

    constexpr size_t Dimension = MeshType::Dimension;

    static_assert(std::is_same_v<TinyVector<Dimension>, InputType>, "invalid input data type");
    static_assert(std::is_same_v<std::remove_const_t<typename OutputArrayT::data_type>, OutputType>,
                  "invalid output data type");

    auto& expression    = Adapter::getFunctionExpression(function_symbol_id);
    auto convert_result = Adapter::getResultConverter(expression.m_data_type);

    auto context_list = Adapter::getContextList(expression);

    using execution_space = typename Kokkos::DefaultExecutionSpace::execution_space;
    Kokkos::Experimental::UniqueToken<execution_space, Kokkos::Experimental::UniqueTokenScope::Global> tokens;

    if constexpr (std::is_arithmetic_v<OutputType>) {
      value.fill(0);
    } else if constexpr (is_tiny_vector_v<OutputType> or is_tiny_matrix_v<OutputType>) {
      value.fill(zero);
    } else {
      static_assert(std::is_same_v<OutputType, double>, "unexpected output type");
    }

    const auto& connectivity = mesh.connectivity();

    const auto cell_to_node_matrix = connectivity.cellToNodeMatrix();
    const auto cell_type           = connectivity.cellType();
    const auto xr                  = mesh.xr();

    const auto cell_to_face_matrix = connectivity.cellToFaceMatrix();
    FaceArray<const Rd> xl;
    if constexpr (is_polynomial_mesh_v<MeshType>) {
      xl = mesh.xl();
    }

    const auto xj = MeshDataManager::instance().getMeshData(mesh).xj();

    auto invalid_cell = std::make_pair(false, CellId{});

    const auto qf = [&] {
      if constexpr (Dimension == 1) {
        return QuadratureManager::instance().getLineFormula(quadrature_descriptor);
      } else if constexpr (Dimension == 2) {
        return QuadratureManager::instance().getSquareFormula(quadrature_descriptor);
      } else {
        static_assert(Dimension == 3);
        return QuadratureManager::instance().getCubeFormula(quadrature_descriptor);
      }
    }();

    parallel_for(cell_list.size(), [=, &mesh, &connectivity, &expression, &tokens, &invalid_cell, &qf,
                                    &cell_list](typename ListTypeT::index_type index) {
      const int32_t t        = tokens.acquire();
      auto& execution_policy = context_list[t];

      const CellId cell_id = cell_list.cellId(index);

      const auto cell_to_node = cell_to_node_matrix[cell_id];

      if constexpr (Dimension == 1) {
        switch (cell_type[cell_id]) {
        case CellType::Line: {
          const LineTransformation<1> T(xr[cell_to_node[0]], xr[cell_to_node[1]]);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant() * convert_result(std::move(result));
          }
          break;
        }
          // LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP
        }
      } else if constexpr (Dimension == 2) {
        switch (cell_type[cell_id]) {
        case CellType::Triangle: {
          if constexpr (is_polygonal_mesh_v<MeshType>) {
            const SquareTransformation<2> T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],
                                            xr[cell_to_node[2]]);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
            }
          } else {
            static_assert(is_polynomial_mesh_v<MeshType>);
            if (mesh.degree() == 2) {
              const auto cell_to_face = cell_to_face_matrix[cell_id];

              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a01 = xl[cell_to_face[1]][0];
              const Rd& a02 = xr[cell_to_node[2]];

              const Rd& a10 = xl[cell_to_face[2]][0];
              const Rd& a11 = xj[cell_id];
              const Rd& a20 = xr[cell_to_node[1]];

              const Rd& a21 = xl[cell_to_face[0]][0];
              const Rd& a12 = xr[cell_to_node[2]];
              const Rd& a22 = xr[cell_to_node[2]];

              const SquareQ2Transformation<2> T(a00, a01, a02,   //
                                                a10, a11, a12,   //
                                                a20, a21, a22);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            } else if (mesh.degree() == 3) {
              const auto cell_to_face = cell_to_face_matrix[cell_id];
              FaceValuePerCell<const bool> cell_face_is_reversed;

              if constexpr (Dimension > 1) {
                cell_face_is_reversed = connectivity.cellFaceIsReversed();
              }

              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a30 = xr[cell_to_node[1]];
              const Rd& a33 = xr[cell_to_node[2]];
              const Rd& a03 = xr[cell_to_node[2]];

              const Rd& a23 = xr[cell_to_node[2]];
              const Rd& a13 = xr[cell_to_node[2]];

              const size_t face01_node0 = (cell_face_is_reversed[cell_id][2]) ? 1 : 0;

              const Rd& a10 = xl[cell_to_face[2]][face01_node0];
              const Rd& a20 = xl[cell_to_face[2]][1 - face01_node0];

              const size_t face12_node0 = (cell_face_is_reversed[cell_id][0]) ? 1 : 0;

              const Rd& a31 = xl[cell_to_face[0]][face12_node0];
              const Rd& a32 = xl[cell_to_face[0]][1 - face12_node0];

              const size_t face30_node0 = (cell_face_is_reversed[cell_id][1]) ? 1 : 0;

              const Rd& a02 = xl[cell_to_face[1]][face30_node0];
              const Rd& a01 = xl[cell_to_face[1]][1 - face30_node0];

              const Rd& a11 = 1. / 3 * xr[cell_to_node[0]] + 2. / 3 * xj[cell_id];
              const Rd& a21 = 1. / 3 * xr[cell_to_node[1]] + 2. / 3 * xj[cell_id];

              const Rd& a12 = 2. / 3 * a02 + 1. / 3 * a32;
              const Rd& a22 = 1. / 3 * a02 + 2. / 3 * a32;

              const SquareQ3Transformation<2> T(a00, a01, a02, a03, a10, a11, a12, a13, a20, a21, a22, a23, a30, a31,
                                                a32, a33);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }

            } else {
              throw NotImplementedError("IntegratedOnCells only for mesh_degree<4");
            }
          }
          break;
        }   // triangle
        case CellType::Quadrangle: {
          if constexpr (is_polygonal_mesh_v<MeshType>) {
            const SquareTransformation<2> T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],
                                            xr[cell_to_node[3]]);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
            }
          } else {
            static_assert(is_polynomial_mesh_v<MeshType>);

            const auto cell_to_face = cell_to_face_matrix[cell_id];

            FaceValuePerCell<const bool> cell_face_is_reversed;
            if constexpr (Dimension > 1) {
              cell_face_is_reversed = connectivity.cellFaceIsReversed();
            }

            if (mesh.degree() == 2) {
              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a01 = xl[cell_to_face[3]][0];
              const Rd& a02 = xr[cell_to_node[3]];

              const Rd& a10 = xl[cell_to_face[0]][0];
              const Rd& a11 = xj[cell_id];
              const Rd& a20 = xr[cell_to_node[1]];

              const Rd& a21 = xl[cell_to_face[1]][0];
              const Rd& a12 = xl[cell_to_face[2]][0];
              const Rd& a22 = xr[cell_to_node[2]];

              const SquareQ2Transformation<2> T(a00, a01, a02,   //
                                                a10, a11, a12,   //
                                                a20, a21, a22);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            } else if (mesh.degree() == 3) {
              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a30 = xr[cell_to_node[1]];
              const Rd& a33 = xr[cell_to_node[2]];
              const Rd& a03 = xr[cell_to_node[3]];

              const size_t face01_node0 = (cell_face_is_reversed[cell_id][0]) ? 1 : 0;

              const Rd& a10 = xl[cell_to_face[0]][face01_node0];
              const Rd& a20 = xl[cell_to_face[0]][1 - face01_node0];

              const size_t face12_node0 = (cell_face_is_reversed[cell_id][1]) ? 1 : 0;

              const Rd& a31 = xl[cell_to_face[1]][face12_node0];
              const Rd& a32 = xl[cell_to_face[1]][1 - face12_node0];

              const size_t face23_node0 = (cell_face_is_reversed[cell_id][2]) ? 1 : 0;

              const Rd& a23 = xl[cell_to_face[2]][face23_node0];
              const Rd& a13 = xl[cell_to_face[2]][1 - face23_node0];

              const size_t face30_node0 = (cell_face_is_reversed[cell_id][3]) ? 1 : 0;

              const Rd& a02 = xl[cell_to_face[3]][face30_node0];
              const Rd& a01 = xl[cell_to_face[3]][1 - face30_node0];

              const Rd& a11 = 1. / 3 * xr[cell_to_node[0]] + 2. / 3 * xj[cell_id];
              const Rd& a21 = 1. / 3 * xr[cell_to_node[1]] + 2. / 3 * xj[cell_id];

              const Rd& a12 = 1. / 3 * xr[cell_to_node[3]] + 2. / 3 * xj[cell_id];
              const Rd& a22 = 1. / 3 * xr[cell_to_node[2]] + 2. / 3 * xj[cell_id];

              const SquareQ3Transformation<2> T(a00, a01, a02, a03, a10, a11, a12, a13, a20, a21, a22, a23, a30, a31,
                                                a32, a33);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }

            } else {
              throw NotImplementedError("IntegratedOnCells only for mesh_degree<4");
            }
          }
          break;
        }   // quad
            //  LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP
        }   // switch
      } else {
        static_assert(Dimension == 3);

        switch (cell_type[cell_id]) {
        case CellType::Tetrahedron: {
          const CubeTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],
                                     xr[cell_to_node[2]],   //
                                     xr[cell_to_node[3]], xr[cell_to_node[3]], xr[cell_to_node[3]],
                                     xr[cell_to_node[3]]);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
          }
          break;
        }
        case CellType::Pyramid: {
          if (cell_to_node.size() == 5) {
            const CubeTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],   //
                                       xr[cell_to_node[3]],                                             //
                                       xr[cell_to_node[4]], xr[cell_to_node[4]], xr[cell_to_node[4]],
                                       xr[cell_to_node[4]]);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
            }
          } else {
            // LCOV_EXCL_START
            throw NotImplementedError("integration on pyramid with non-quadrangular base (number of nodes " +
                                      stringify(cell_to_node.size()) + ")");
            // LCOV_EXCL_STOP
          }
          break;
        }
        case CellType::Diamond: {
          if (cell_to_node.size() == 5) {
            {   // top tetrahedron
              const CubeTransformation T0(xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],
                                          xr[cell_to_node[3]],   //
                                          xr[cell_to_node[4]], xr[cell_to_node[4]], xr[cell_to_node[4]],
                                          xr[cell_to_node[4]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T0(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T0.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
            {   // bottom tetrahedron
              const CubeTransformation T1(xr[cell_to_node[3]], xr[cell_to_node[3]], xr[cell_to_node[2]],
                                          xr[cell_to_node[1]],   //
                                          xr[cell_to_node[0]], xr[cell_to_node[0]], xr[cell_to_node[0]],
                                          xr[cell_to_node[0]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T1(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T1.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
          } else if (cell_to_node.size() == 6) {
            {   // top pyramid
              const CubeTransformation T0(xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],
                                          xr[cell_to_node[4]],   //
                                          xr[cell_to_node[5]], xr[cell_to_node[5]], xr[cell_to_node[5]],
                                          xr[cell_to_node[5]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T0(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T0.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
            {   // bottom pyramid
              const CubeTransformation T1(xr[cell_to_node[4]], xr[cell_to_node[3]], xr[cell_to_node[2]],
                                          xr[cell_to_node[1]],   //
                                          xr[cell_to_node[0]], xr[cell_to_node[0]], xr[cell_to_node[0]],
                                          xr[cell_to_node[0]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T1(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T1.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
          } else {
            // LCOV_EXCL_START
            throw NotImplementedError("integration on diamond with non-quadrangular base (" +
                                      stringify(cell_to_node.size()) + ")");
            // LCOV_EXCL_STOP
          }
          break;
        }
        case CellType::Prism: {
          const CubeTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]],   //
                                     xr[cell_to_node[2]], xr[cell_to_node[2]],   //
                                     xr[cell_to_node[3]], xr[cell_to_node[4]],   //
                                     xr[cell_to_node[5]], xr[cell_to_node[5]]);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
          }
          break;
        }
        case CellType::Hexahedron: {
          const CubeTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],
                                     xr[cell_to_node[4]], xr[cell_to_node[5]], xr[cell_to_node[6]],
                                     xr[cell_to_node[7]]);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
          }
          break;
        }
          // LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP
        }   // switch
      }     // else
      tokens.release(t);
    });

    // LCOV_EXCL_START
    if (invalid_cell.first) {
      std::ostringstream os;
      os << "invalid cell type for integration: " << name(cell_type[invalid_cell.second]);
      throw UnexpectedError(os.str());
    }
    // LCOV_EXCL_STOP
  }

  template <MeshConcept MeshType, typename OutputArrayT, typename ListTypeT>
  static PUGS_INLINE void
  _directIntegrateTo(const FunctionSymbolId& function_symbol_id,
                     const IQuadratureDescriptor& quadrature_descriptor,
                     const MeshType& mesh,
                     const ListTypeT& cell_list,
                     OutputArrayT& value)
  {
    static_assert((is_polygonal_mesh_v<MeshType> or is_polynomial_mesh_v<MeshType>)   //
                  and(MeshType::Dimension == InputType::Dimension));

    static_assert(is_polygonal_mesh_v<MeshType> or (is_polynomial_mesh_v<MeshType> and (MeshType::Dimension == 2)),
                  "Polynomial meshes are only supported in dimension 2");

    if constexpr (is_polynomial_mesh_v<MeshType>) {
      if (mesh.degree() > 3) {
        throw NotImplementedError("IntegrateOnCells only support polynomial meshes of degree <= 3");
      }
    }

    Assert(not quadrature_descriptor.isTensorial());

    constexpr size_t Dimension = MeshType::Dimension;

    static_assert(std::is_same_v<TinyVector<Dimension>, InputType>, "invalid input data type");
    static_assert(std::is_same_v<std::remove_const_t<typename OutputArrayT::data_type>, OutputType>,
                  "invalid output data type");

    auto& expression    = Adapter::getFunctionExpression(function_symbol_id);
    auto convert_result = Adapter::getResultConverter(expression.m_data_type);

    auto context_list = Adapter::getContextList(expression);

    using execution_space = typename Kokkos::DefaultExecutionSpace::execution_space;
    Kokkos::Experimental::UniqueToken<execution_space, Kokkos::Experimental::UniqueTokenScope::Global> tokens;

    if constexpr (std::is_arithmetic_v<OutputType>) {
      value.fill(0);
    } else if constexpr (is_tiny_vector_v<OutputType> or is_tiny_matrix_v<OutputType>) {
      value.fill(zero);
    } else {
      static_assert(std::is_same_v<OutputType, double>, "unexpected output type");
    }

    const auto& connectivity = mesh.connectivity();

    const auto cell_to_node_matrix = connectivity.cellToNodeMatrix();
    const auto cell_type           = connectivity.cellType();

    FaceValuePerCell<const bool> cell_face_is_reversed;
    if constexpr (Dimension > 1) {
      cell_face_is_reversed = connectivity.cellFaceIsReversed();
    }
    const auto xr = mesh.xr();

    const auto cell_to_face_matrix = connectivity.cellToFaceMatrix();
    FaceArray<const Rd> xl;
    if constexpr (is_polynomial_mesh_v<MeshType>) {
      xl = mesh.xl();
    }

    const auto xj = MeshDataManager::instance().getMeshData(mesh).xj();

    auto invalid_cell = std::make_pair(false, CellId{});

    parallel_for(cell_list.size(), [=, &mesh, &expression, &tokens, &invalid_cell, &quadrature_descriptor,
                                    &cell_list](const typename ListTypeT::index_type& index) {
      const int32_t t        = tokens.acquire();
      auto& execution_policy = context_list[t];

      const CellId cell_id = cell_list.cellId(index);

      const auto cell_to_node = cell_to_node_matrix[cell_id];

      if constexpr (Dimension == 1) {
        switch (cell_type[cell_id]) {
        case CellType::Line: {
          const LineTransformation<1> T(xr[cell_to_node[0]], xr[cell_to_node[1]]);
          const auto qf = QuadratureManager::instance().getLineFormula(quadrature_descriptor);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant() * convert_result(std::move(result));
          }
          break;
        }
          // LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP
        }
      } else if constexpr (Dimension == 2) {
        switch (cell_type[cell_id]) {
        case CellType::Triangle: {
          if constexpr (is_polygonal_mesh_v<MeshType>) {
            const TriangleTransformation<2> T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]]);
            const auto qf = QuadratureManager::instance().getTriangleFormula(quadrature_descriptor);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant() * convert_result(std::move(result));
            }
          } else {
            static_assert(is_polynomial_mesh_v<MeshType>);

            const auto cell_to_face = cell_to_face_matrix[cell_id];
            const auto qf           = QuadratureManager::instance().getTriangleFormula(quadrature_descriptor);

            if (mesh.degree() == 2) {
              const Rd& a0 = xr[cell_to_node[0]];
              const Rd& a1 = xr[cell_to_node[1]];
              const Rd& a2 = xr[cell_to_node[2]];

              const Rd& a01 = xl[cell_to_face[2]][0];
              const Rd& a12 = xl[cell_to_face[0]][0];
              const Rd& a20 = xl[cell_to_face[1]][0];

              const TriangleP2Transformation<2> T(a0, a1, a2,   //
                                                  a12, a20, a01);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            } else if (mesh.degree() == 3) {
              const Rd& a0 = xr[cell_to_node[0]];
              const Rd& a1 = xr[cell_to_node[1]];
              const Rd& a2 = xr[cell_to_node[2]];

              const size_t face01_node0 = (cell_face_is_reversed[cell_id][2]) ? 1 : 0;

              const Rd& a001 = xl[cell_to_face[2]][face01_node0];
              const Rd& a011 = xl[cell_to_face[2]][1 - face01_node0];

              const size_t face12_node0 = (cell_face_is_reversed[cell_id][0]) ? 1 : 0;

              const Rd& a112 = xl[cell_to_face[0]][face12_node0];
              const Rd& a122 = xl[cell_to_face[0]][1 - face12_node0];

              const size_t face20_node0 = (cell_face_is_reversed[cell_id][1]) ? 1 : 0;

              const Rd& a220 = xl[cell_to_face[1]][face20_node0];
              const Rd& a200 = xl[cell_to_face[1]][1 - face20_node0];

              const Rd& a012 = 1. / 3 * (a0 + a1 + a2);

              const TriangleP3Transformation<2> T(a0, a1, a2, a112, a122, a220, a200, a001, a011, a012);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            } else {
              throw NotImplementedError("IntegratedOnCells only for mesh_degree<4");
            }
          }
          break;
        }
        case CellType::Quadrangle: {
          if constexpr (is_polygonal_mesh_v<MeshType>) {
            const SquareTransformation<2> T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],
                                            xr[cell_to_node[3]]);
            const auto qf = QuadratureManager::instance().getSquareFormula(quadrature_descriptor);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
            }
          } else {
            static_assert(is_polynomial_mesh_v<MeshType>);

            const auto cell_to_face = cell_to_face_matrix[cell_id];
            if (mesh.degree() == 2) {
              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a01 = xl[cell_to_face[3]][0];
              const Rd& a02 = xr[cell_to_node[3]];

              const Rd& a10 = xl[cell_to_face[0]][0];
              const Rd& a11 = xj[cell_id];
              const Rd& a12 = xl[cell_to_face[2]][0];

              const Rd& a20 = xr[cell_to_node[1]];
              const Rd& a21 = xl[cell_to_face[1]][0];
              const Rd& a22 = xr[cell_to_node[2]];

              const SquareQ2Transformation<2> T(a00, a01, a02,   //
                                                a10, a11, a12,   //
                                                a20, a21, a22);
              const auto qf = QuadratureManager::instance().getSquareFormula(quadrature_descriptor);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }

            } else if (mesh.degree() == 3) {
              const Rd& a00 = xr[cell_to_node[0]];
              const Rd& a30 = xr[cell_to_node[1]];
              const Rd& a33 = xr[cell_to_node[2]];
              const Rd& a03 = xr[cell_to_node[3]];

              const size_t face01_node0 = (cell_face_is_reversed[cell_id][0]) ? 1 : 0;

              const Rd& a10 = xl[cell_to_face[0]][face01_node0];
              const Rd& a20 = xl[cell_to_face[0]][1 - face01_node0];

              const size_t face12_node0 = (cell_face_is_reversed[cell_id][1]) ? 1 : 0;

              const Rd& a31 = xl[cell_to_face[1]][face12_node0];
              const Rd& a32 = xl[cell_to_face[1]][1 - face12_node0];

              const size_t face23_node0 = (cell_face_is_reversed[cell_id][2]) ? 1 : 0;

              const Rd& a23 = xl[cell_to_face[2]][face23_node0];
              const Rd& a13 = xl[cell_to_face[2]][1 - face23_node0];

              const size_t face30_node0 = (cell_face_is_reversed[cell_id][3]) ? 1 : 0;

              const Rd& a02 = xl[cell_to_face[3]][face30_node0];
              const Rd& a01 = xl[cell_to_face[3]][1 - face30_node0];

              const Rd& a11 = 1. / 3 * xr[cell_to_node[0]] + 2. / 3 * xj[cell_id];
              const Rd& a21 = 1. / 3 * xr[cell_to_node[1]] + 2. / 3 * xj[cell_id];

              const Rd& a12 = 1. / 3 * xr[cell_to_node[3]] + 2. / 3 * xj[cell_id];
              const Rd& a22 = 1. / 3 * xr[cell_to_node[2]] + 2. / 3 * xj[cell_id];

              const SquareQ3Transformation<2> T(a00, a01, a02, a03, a10, a11, a12, a13, a20, a21, a22, a23, a30, a31,
                                                a32, a33);
              const auto qf = QuadratureManager::instance().getSquareFormula(quadrature_descriptor);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            } else {
              throw NotImplementedError("IntegratedOnCells only for mesh_degree<4");
            }
          }
          break;
        }   // case quad
            //  LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP

        }   // switch
      } else {
        static_assert(Dimension == 3);

        switch (cell_type[cell_id]) {
        case CellType::Tetrahedron: {
          const TetrahedronTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]],   //
                                            xr[cell_to_node[2]], xr[cell_to_node[3]]);
          const auto qf = QuadratureManager::instance().getTetrahedronFormula(quadrature_descriptor);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant() * convert_result(std::move(result));
          }
          break;
        }
        case CellType::Pyramid: {
          if (cell_to_node.size() == 5) {
            const PyramidTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],   //
                                          xr[cell_to_node[3]], xr[cell_to_node[4]]);
            const auto qf = QuadratureManager::instance().getPyramidFormula(quadrature_descriptor);

            for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
              const auto xi = qf.point(i_point);
              Adapter::convertArgs(execution_policy.currentContext(), T(xi));
              auto result = expression.execute(execution_policy);
              value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
            }
          } else {
            // LCOV_EXCL_START
            throw NotImplementedError("integration on pyramid with non-quadrangular base");
            // LCOV_EXCL_STOP
          }
          break;
        }
        case CellType::Diamond: {
          if (cell_to_node.size() == 5) {
            const auto qf = QuadratureManager::instance().getTetrahedronFormula(quadrature_descriptor);
            {   // top tetrahedron
              const TetrahedronTransformation T0(xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],   //
                                                 xr[cell_to_node[4]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T0(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T0.jacobianDeterminant() * convert_result(std::move(result));
              }
            }
            {   // bottom tetrahedron
              const TetrahedronTransformation T1(xr[cell_to_node[3]], xr[cell_to_node[2]], xr[cell_to_node[1]],   //
                                                 xr[cell_to_node[0]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T1(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T1.jacobianDeterminant() * convert_result(std::move(result));
              }
            }
          } else if (cell_to_node.size() == 6) {
            const auto qf = QuadratureManager::instance().getPyramidFormula(quadrature_descriptor);
            {   // top pyramid
              const PyramidTransformation T0(xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],
                                             xr[cell_to_node[4]], xr[cell_to_node[5]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T0(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T0.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
            {   // bottom pyramid
              const PyramidTransformation T1(xr[cell_to_node[4]], xr[cell_to_node[3]], xr[cell_to_node[2]],
                                             xr[cell_to_node[1]], xr[cell_to_node[0]]);

              for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
                const auto xi = qf.point(i_point);
                Adapter::convertArgs(execution_policy.currentContext(), T1(xi));
                auto result = expression.execute(execution_policy);
                value[index] += qf.weight(i_point) * T1.jacobianDeterminant(xi) * convert_result(std::move(result));
              }
            }
          } else {
            // LCOV_EXCL_START
            throw NotImplementedError("integration on diamond with non-quadrangular base (" +
                                      stringify(cell_to_node.size()) + ")");
            // LCOV_EXCL_STOP
          }
          break;
        }
        case CellType::Prism: {
          const PrismTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]],
                                      xr[cell_to_node[3]], xr[cell_to_node[4]], xr[cell_to_node[5]]);
          const auto qf = QuadratureManager::instance().getPrismFormula(quadrature_descriptor);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
          }
          break;
        }
        case CellType::Hexahedron: {
          const CubeTransformation T(xr[cell_to_node[0]], xr[cell_to_node[1]], xr[cell_to_node[2]], xr[cell_to_node[3]],
                                     xr[cell_to_node[4]], xr[cell_to_node[5]], xr[cell_to_node[6]],
                                     xr[cell_to_node[7]]);
          const auto qf = QuadratureManager::instance().getCubeFormula(quadrature_descriptor);

          for (size_t i_point = 0; i_point < qf.numberOfPoints(); ++i_point) {
            const auto xi = qf.point(i_point);
            Adapter::convertArgs(execution_policy.currentContext(), T(xi));
            auto result = expression.execute(execution_policy);
            value[index] += qf.weight(i_point) * T.jacobianDeterminant(xi) * convert_result(std::move(result));
          }
          break;
        }
          // LCOV_EXCL_START
        default: {
          invalid_cell = std::make_pair(true, cell_id);
          break;
        }
          // LCOV_EXCL_STOP
        }   // switch
      }     // fin test dimension
      tokens.release(t);
    });

    // LCOV_EXCL_START
    if (invalid_cell.first) {
      std::ostringstream os;
      os << "invalid cell type for integration: " << name(cell_type[invalid_cell.second]);
      throw UnexpectedError(os.str());
    }
    // LCOV_EXCL_STOP
  }

 public:
  template <MeshConcept MeshType, typename ArrayT>
  static PUGS_INLINE void
  integrateTo(const FunctionSymbolId& function_symbol_id,
              const IQuadratureDescriptor& quadrature_descriptor,
              const MeshType& mesh,
              ArrayT& value)
  {
    constexpr size_t Dimension = MeshType::Dimension;

    if (quadrature_descriptor.isTensorial()) {
      _tensorialIntegrateTo<MeshType, ArrayT>(function_symbol_id, quadrature_descriptor, mesh,
                                              AllCells<Dimension>{mesh.connectivity()}, value);
    } else {
      _directIntegrateTo<MeshType, ArrayT>(function_symbol_id, quadrature_descriptor, mesh,
                                           AllCells<Dimension>{mesh.connectivity()}, value);
    }
  }

  template <MeshConcept MeshType, template <typename DataType> typename ArrayT>
  static PUGS_INLINE ArrayT<OutputType>
  integrate(const FunctionSymbolId& function_symbol_id,
            const IQuadratureDescriptor& quadrature_descriptor,
            const MeshType& mesh,
            const ArrayT<const CellId>& cell_list)
  {
    static_assert((is_polygonal_mesh_v<MeshType> or is_polynomial_mesh_v<MeshType>)   //
                  and(MeshType::Dimension == InputType::Dimension));
    ArrayT<OutputType> value(size(cell_list));
    if (quadrature_descriptor.isTensorial()) {
      _tensorialIntegrateTo<MeshType, ArrayT<OutputType>>(function_symbol_id, quadrature_descriptor, mesh,
                                                          CellList{cell_list}, value);
    } else {
      _directIntegrateTo<MeshType, ArrayT<OutputType>>(function_symbol_id, quadrature_descriptor, mesh,
                                                       CellList{cell_list}, value);
    }

    return value;
  }

  template <MeshConcept MeshType, template <typename DataType> typename ArrayT>
  static PUGS_INLINE ArrayT<OutputType>
  integrate(const FunctionSymbolId& function_symbol_id,
            const IQuadratureDescriptor& quadrature_descriptor,
            const MeshType& mesh,
            const ArrayT<CellId>& cell_list)
  {
    return integrate(function_symbol_id, quadrature_descriptor, mesh, ArrayT<const CellId>{cell_list});
  }
};

#endif   // INTEGRATE_ON_CELLS_HPP
