#include <mesh/MeshData.hpp>

#include <geometry/LineCubicTransformation.hpp>
#include <geometry/LineParabolicTransformation.hpp>
#include <geometry/LineTransformation.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ItemValueVariant.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/PolynomialMesh.hpp>
#include <output/NamedItemValueVariant.hpp>
#include <output/VTKWriter.hpp>
#include <utils/Exceptions.hpp>

template <size_t Dimension>
void
MeshData<Mesh<Dimension>>::_storeBadMesh()
{
  VTKWriter writer("bad_mesh");
  writer.writeOnMesh(std::make_shared<MeshVariant>(
                       std::make_shared<const Mesh<Dimension>>(m_mesh.shared_connectivity(), m_mesh.xr())),
                     {std::make_shared<NamedItemValueVariant>(std::make_shared<ItemValueVariant>(m_Vj), "volume")});
  std::ostringstream error_msg;
  error_msg << "mesh contains cells of non-positive volume (see " << rang::fgB::yellow << "bad_mesh.pvd"
            << rang::fg::reset << " file).";
  throw NormalError(error_msg.str());
}

template void MeshData<Mesh<1>>::_storeBadMesh();
template void MeshData<Mesh<2>>::_storeBadMesh();
template void MeshData<Mesh<3>>::_storeBadMesh();
void
MeshData<PolynomialMesh<2>>::_storeBadMesh()
{
  VTKWriter writer("bad_mesh");
  writer.writeOnMesh(m_mesh.meshVariant(),
                     {std::make_shared<NamedItemValueVariant>(std::make_shared<ItemValueVariant>(m_Vj), "volume")});
  std::ostringstream error_msg;
  error_msg << "mesh contains cells of non-positive volume (see " << rang::fgB::yellow << "bad_mesh.pvd"
            << rang::fg::reset << " file).";
  throw NormalError(error_msg.str());
}

template <size_t MeshDegree>
auto getTransformation(const NodeValue<const TinyVector<2>>& xr,
                       const FaceArray<const TinyVector<2>>& xl,
                       const ItemToItemMatrix<ItemType::face, ItemType::node>& face_to_node_matrix,
                       const FaceId face_id);

template <>
auto
getTransformation<1>(const NodeValue<const TinyVector<2>>& xr,
                     const FaceArray<const TinyVector<2>>&,
                     const ItemToItemMatrix<ItemType::face, ItemType::node>& face_to_node_matrix,
                     const FaceId face_id)
{
  return LineTransformation<2>{xr[face_to_node_matrix[face_id][0]],   //
                               xr[face_to_node_matrix[face_id][1]]};
}

template <>
auto
getTransformation<2>(const NodeValue<const TinyVector<2>>& xr,
                     const FaceArray<const TinyVector<2>>& xl,
                     const ItemToItemMatrix<ItemType::face, ItemType::node>& face_to_node_matrix,
                     const FaceId face_id)
{
  return LineParabolicTransformation<2>{xr[face_to_node_matrix[face_id][0]],   //
                                        xl[face_id][0],                        //
                                        xr[face_to_node_matrix[face_id][1]]};
}

template <>
auto
getTransformation<3>(const NodeValue<const TinyVector<2>>& xr,
                     const FaceArray<const TinyVector<2>>& xl,
                     const ItemToItemMatrix<ItemType::face, ItemType::node>& face_to_node_matrix,
                     const FaceId face_id)
{
  return LineCubicTransformation<2>{xr[face_to_node_matrix[face_id][0]],   //
                                    xl[face_id][0],                        //
                                    xl[face_id][1],                        //
                                    xr[face_to_node_matrix[face_id][1]]};
}

template <size_t MeshDegree>
void
MeshData<PolynomialMesh<2>>::_computeCellCentroidForMeshDegree()
{
  FaceValue<Rd> face_integral{m_mesh.connectivity()};

  using R1 = TinyVector<1>;

  const auto xr = m_mesh.xr();
  const auto xl = m_mesh.xl();

  auto face_to_node_matrix = m_mesh.connectivity().faceToNodeMatrix();

  constexpr size_t total_degree = 3 * MeshDegree - 1;

  QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(total_degree));
  parallel_for(
    m_mesh.numberOfFaces(), PUGS_LAMBDA(const FaceId face_id) {
      auto T = getTransformation<MeshDegree>(xr, xl, face_to_node_matrix, face_id);

      Rd integral = zero;
      for (size_t i_ksi = 0; i_ksi < qf.numberOfPoints(); ++i_ksi) {
        const R1& ksi    = qf.point(i_ksi);
        const Rd x       = T(ksi);
        const Rd x_prime = T.velocity(ksi);
        const Rd x_prime_perp{x_prime[1], -x_prime[0]};

        integral += qf.weight(i_ksi) * Rd{x[0] * x[0] * (x_prime_perp[0]),   //
                                          x[1] * x[1] * (x_prime_perp[1])};
      }

      face_integral[face_id] = integral;
    });

  auto cell_face_is_reversed = m_mesh.connectivity().cellFaceIsReversed();
  auto cell_to_face_matrix   = m_mesh.connectivity().cellToFaceMatrix();

  CellValue<const double> Vj = this->Vj();
  CellValue<Rd> cell_centroid{m_mesh.connectivity()};
  parallel_for(
    m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
      auto cell_faces = cell_to_face_matrix[cell_id];
      Rd xj           = zero;
      for (size_t i_face = 0; i_face < cell_faces.size(); ++i_face) {
        const FaceId face_id     = cell_faces[i_face];
        const double orientation = (cell_face_is_reversed[cell_id][i_face]) ? -1 : 1;

        xj += orientation * face_integral[face_id];
      }
      cell_centroid[cell_id] = 1 / (2 * Vj[cell_id]) * xj;
    });

  m_cell_centroid = cell_centroid;
}

void
MeshData<PolynomialMesh<2>>::_computeCellCentroid()
{
  switch (m_mesh.degree()) {
  case 1: {
    this->_computeCellCentroidForMeshDegree<1>();
    break;
  }
  case 2: {
    this->_computeCellCentroidForMeshDegree<2>();
    break;
  }
  case 3: {
    this->_computeCellCentroidForMeshDegree<3>();
    break;
  }
  default: {
    throw NotImplementedError("mesh degree > 3");
  }
  }
}
template <size_t MeshDegree>
void
MeshData<PolynomialMesh<2>>::_computeCellVolumeForMeshDegree()
{
  FaceValue<double> face_integral{m_mesh.connectivity()};

  using R1 = TinyVector<1>;

  const auto xr = m_mesh.xr();
  const auto xl = m_mesh.xl();

  auto face_to_node_matrix = m_mesh.connectivity().faceToNodeMatrix();

  constexpr size_t total_degree = 2 * MeshDegree - 1;

  QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(total_degree));
  parallel_for(
    m_mesh.numberOfFaces(), PUGS_LAMBDA(const FaceId face_id) {
      auto T = getTransformation<MeshDegree>(xr, xl, face_to_node_matrix, face_id);

      double integral = 0;
      for (size_t i_ksi = 0; i_ksi < qf.numberOfPoints(); ++i_ksi) {
        const R1& ksi    = qf.point(i_ksi);
        const Rd x       = T(ksi);
        const Rd x_prime = T.velocity(ksi);
        const Rd x_prime_perp{x_prime[1], -x_prime[0]};

        integral += qf.weight(i_ksi) * dot(x, x_prime_perp);
      }

      face_integral[face_id] = inv_Dimension * integral;
    });

  auto cell_face_is_reversed = m_mesh.connectivity().cellFaceIsReversed();
  auto cell_to_face_matrix   = m_mesh.connectivity().cellToFaceMatrix();

  CellValue<double> Vj{m_mesh.connectivity()};
  parallel_for(
    m_mesh.numberOfCells(), PUGS_LAMBDA(const CellId cell_id) {
      auto cell_faces = cell_to_face_matrix[cell_id];
      double volume   = 0;
      for (size_t i_face = 0; i_face < cell_faces.size(); ++i_face) {
        const FaceId face_id     = cell_faces[i_face];
        const double orientation = (cell_face_is_reversed[cell_id][i_face]) ? -1 : 1;

        volume += orientation * face_integral[face_id];
      }
      Vj[cell_id] = volume;
    });

  m_Vj = Vj;
}

void
MeshData<PolynomialMesh<2>>::_computeCellVolume()
{
  switch (m_mesh.degree()) {
  case 1: {
    this->_computeCellVolumeForMeshDegree<1>();
    break;
  }
  case 2: {
    this->_computeCellVolumeForMeshDegree<2>();
    break;
  }
  case 3: {
    this->_computeCellVolumeForMeshDegree<3>();
    break;
  }
  default: {
    throw NotImplementedError("mesh degree > 3");
  }
  }
}
