#include <mesh/MeshFlatFaceBoundary.hpp>

#include <mesh/Connectivity.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshFlatNodeBoundary.hpp>
#include <mesh/MeshNodeBoundaryUtils.hpp>
#include <mesh/PolynomialMesh.hpp>
#include <utils/Messenger.hpp>

template <typename MeshType>
MeshFlatFaceBoundary<MeshType>
getMeshFlatFaceBoundary(const MeshType& mesh, const IBoundaryDescriptor& boundary_descriptor)
{
  MeshFaceBoundary mesh_face_boundary          = getMeshFaceBoundary(mesh, boundary_descriptor);
  MeshFlatNodeBoundary mesh_flat_node_boundary = getMeshFlatNodeBoundary(mesh, boundary_descriptor);

  if constexpr (is_polynomial_mesh_v<MeshType>) {
    static_assert(MeshType::Dimension == 2);

    using R2 = TinyVector<2, double>;

    std::array<R2, 2> bounds = getBounds(mesh, mesh_flat_node_boundary.refNodeList());

    const R2& xmin      = bounds[0];
    const R2& xmax      = bounds[1];
    const R2 origin     = 0.5 * (xmin + xmax);
    const double length = l2Norm(xmax - xmin);

    bool is_bad = false;

    const FaceArray<const R2>& xl = mesh.xl();
    const size_t degree           = mesh.degree();

    auto face_to_node_matrix = mesh.connectivity().faceToNodeMatrix();

    auto face_list = mesh_face_boundary.faceList();
    parallel_for(face_list.size(), [=, &is_bad](int i_face) {
      const FaceId face_id = face_list[i_face];

      auto face_nodes = face_to_node_matrix[face_id];

      auto inner_face_vertices = xl[face_id];
      for (size_t inner_node = 0; inner_node < degree - 1; ++inner_node) {
        const R2& x = inner_face_vertices[inner_node];
        if (std::abs(dot(x - origin, mesh_flat_node_boundary.outgoingNormal())) > 1E-13 * length) {
          is_bad = true;
        }
      }
    });

    if (parallel::allReduceOr(is_bad)) {
      std::ostringstream ost;
      ost << "invalid boundary \"" << rang::fgB::yellow << mesh_face_boundary.refFaceList().refId()
          << rang::style::reset << "\": boundary is not flat!";
      throw NormalError(ost.str());
    }
  }

  return MeshFlatFaceBoundary<MeshType>{mesh, mesh_face_boundary.refFaceList(), mesh_flat_node_boundary.origin(),
                                        mesh_flat_node_boundary.outgoingNormal()};
}

template MeshFlatFaceBoundary<Mesh<1>> getMeshFlatFaceBoundary(const Mesh<1>&, const IBoundaryDescriptor&);
template MeshFlatFaceBoundary<Mesh<2>> getMeshFlatFaceBoundary(const Mesh<2>&, const IBoundaryDescriptor&);
template MeshFlatFaceBoundary<Mesh<3>> getMeshFlatFaceBoundary(const Mesh<3>&, const IBoundaryDescriptor&);

template MeshFlatFaceBoundary<PolynomialMesh<2>> getMeshFlatFaceBoundary(const PolynomialMesh<2>&,
                                                                         const IBoundaryDescriptor&);
