#ifndef POLYNOMIAL_MESH_HPP
#define POLYNOMIAL_MESH_HPP

#include <algebra/TinyVector.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ItemArray.hpp>
#include <mesh/ItemValue.hpp>
#include <mesh/MeshVariant.hpp>
#include <utils/GlobalVariableManager.hpp>

#include <memory>

template <size_t MeshDimension>
class PolynomialMesh : public std::enable_shared_from_this<PolynomialMesh<MeshDimension>>
{
 public:
  static constexpr size_t Dimension = MeshDimension;
  using Connectivity                = ::Connectivity<Dimension>;

  using Rd = TinyVector<Dimension>;

 private:
  const size_t m_id;

  const std::shared_ptr<const Connectivity> m_connectivity;
  const NodeValue<const Rd> m_xr;
  const FaceArray<const Rd> m_xl;

  std::ostream&
  _write(std::ostream& os) const
  {
    return os << *m_connectivity;
  }

 public:
  friend std::ostream&
  operator<<(std::ostream& os, const PolynomialMesh& mesh)
  {
    return mesh._write(os);
  }

  PUGS_INLINE
  size_t
  id() const
  {
    return m_id;
  }

  PUGS_INLINE
  std::shared_ptr<const PolynomialMesh<Dimension>>
  shared_ptr() const
  {
    return this->shared_from_this();
  }

  PUGS_INLINE
  std::shared_ptr<const MeshVariant>
  meshVariant() const
  {
    return std::make_shared<const MeshVariant>(this->shared_ptr());
  }

  PUGS_INLINE
  constexpr size_t
  dimension() const
  {
    return Dimension;
  }

  PUGS_INLINE
  size_t
  degree() const
  {
    return 1 + m_xl.sizeOfArrays();
  }

  PUGS_INLINE
  const Connectivity&
  connectivity() const
  {
    return *m_connectivity;
  }

  PUGS_INLINE
  const auto&
  shared_connectivity() const
  {
    return m_connectivity;
  }

  PUGS_INLINE
  size_t
  numberOfNodes() const
  {
    return m_connectivity->numberOfNodes();
  }

  PUGS_INLINE
  size_t
  numberOfEdges() const
  {
    return m_connectivity->numberOfEdges();
  }

  PUGS_INLINE
  size_t
  numberOfFaces() const
  {
    return m_connectivity->numberOfFaces();
  }

  PUGS_INLINE
  size_t
  numberOfCells() const
  {
    return m_connectivity->numberOfCells();
  }

  template <ItemType item_type>
  PUGS_INLINE size_t
  numberOf() const
  {
    return m_connectivity->template numberOf<item_type>();
  }

  PUGS_INLINE
  const NodeValue<const Rd>&
  xr() const
  {
    return m_xr;
  }

  PUGS_INLINE
  const FaceArray<const Rd>&
  xl() const
  {
    return m_xl;
  }

  PolynomialMesh() = delete;

  PUGS_INLINE
  PolynomialMesh(const std::shared_ptr<const Connectivity>& connectivity,
                 const NodeValue<const Rd>& xr,
                 const FaceArray<const Rd>& xl)
    : m_id{GlobalVariableManager::instance().getAndIncrementMeshId()}, m_connectivity{connectivity}, m_xr{xr}, m_xl{xl}
  {
    ;
  }

  PolynomialMesh(const PolynomialMesh&) = delete;
  PolynomialMesh(PolynomialMesh&&)      = delete;

  ~PolynomialMesh();
};

#endif   // POLYNOMIAL_MESH_HPP
