#include <mesh/PolynomialMeshBuilder.hpp>

#include <mesh/Mesh.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/PolynomialMesh.hpp>

#include <string>

template <MeshConcept MeshType>
void
PolynomialMeshBuilder::_build(const MeshType& mesh, const uint64_t& degree)
{
  constexpr size_t Dimension = MeshType::Dimension;
  static_assert(Dimension == 2);

  using Rd = TinyVector<Dimension>;

  if (degree < 1) {
    throw NormalError("provided degree must be at least 1");
  }

  const Connectivity<Dimension>& connectivity = mesh.connectivity();

  const NodeValue<const Rd> xr = mesh.xr();
  FaceArray<Rd> xl(connectivity, degree - 1);

  auto face_to_node_matrix = connectivity.faceToNodeMatrix();

  const size_t number_of_face_nodes = degree - 1;
  const double ratio                = 1. / degree;

  parallel_for(
    connectivity.numberOfFaces(), PUGS_LAMBDA(const FaceId face_id) {
      const Rd x0 = xr[face_to_node_matrix[face_id][0]];
      const Rd x1 = xr[face_to_node_matrix[face_id][1]];

      for (size_t i_face_node = 0; i_face_node < number_of_face_nodes; ++i_face_node) {
        xl[face_id][i_face_node] = (i_face_node + 1) * ratio * x1 + (number_of_face_nodes - i_face_node) * ratio * x0;
      }
    });

  std::shared_ptr<const PolynomialMesh<Dimension>> polynomial_mesh =
    std::make_shared<PolynomialMesh<Dimension>>(mesh.shared_connectivity(), xr, xl);

  m_mesh = std::make_shared<MeshVariant>(polynomial_mesh);
}

PolynomialMeshBuilder::PolynomialMeshBuilder(const std::shared_ptr<const MeshVariant>& mesh_v, const uint64_t& degree)
{
  std::visit(
    [&](auto&& mesh) {
      using MeshType = mesh_type_t<decltype(mesh)>;
      if constexpr (is_polygonal_mesh_v<MeshType>) {
        if constexpr (MeshType::Dimension == 2) {
          this->_build(*mesh, degree);
        } else {
          throw NotImplementedError("cannot build polynomial mesh in dimension " + std::to_string(MeshType::Dimension));
        }
      } else {
        throw NormalError("input mesh must be a polygonal mesh. Provided: " + demangle<MeshType>());
      }
    },
    mesh_v->variant());
}
