#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <language/ast/ASTBuilder.hpp>
#include <language/ast/ASTModulesImporter.hpp>
#include <language/ast/ASTNodeDataTypeBuilder.hpp>
#include <language/ast/ASTNodeExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionEvaluationExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionExpressionBuilder.hpp>
#include <language/ast/ASTNodeTypeCleaner.hpp>
#include <language/ast/ASTSymbolTableBuilder.hpp>
#include <language/utils/PugsFunctionAdapter.hpp>
#include <language/utils/SymbolTable.hpp>

#include <MeshDataBaseForTests.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/DualMeshManager.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshData.hpp>
#include <mesh/MeshDataManager.hpp>
#include <mesh/PolynomialMeshBuilder.hpp>
#include <scheme/CellIntegrator.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/GaussLobattoQuadratureDescriptor.hpp>
#include <analysis/GaussQuadratureDescriptor.hpp>
#include <language/utils/IntegrateOnCells.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("IntegrateOnCells", "[language]")
{
  SECTION("Gauss quadrature")
  {
    auto quadrature_descriptor = GaussQuadratureDescriptor(3);

    SECTION("integrate on all cells")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.numberOfItems(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<double> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R3> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R2x2> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<double> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R3> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R2x2> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<double> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R3> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R2x2> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }

    SECTION("integrate on cell list")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.size(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            Array<CellId> cell_list{mesh_1d->numberOfCells() / 2 + mesh_1d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_1d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            Array<CellId> cell_list{mesh_2d->numberOfCells() / 2 + mesh_2d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_2d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            Array<CellId> cell_list{mesh_3d->numberOfCells() / 2 + mesh_3d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_3d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }
  }

  SECTION("Gauss-Legendre quadrature")
  {
    auto quadrature_descriptor = GaussLegendreQuadratureDescriptor(3);

    SECTION("integrate on all cells")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.numberOfItems(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<double> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R3> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R2x2> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<double> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R3> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R2x2> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<double> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R3> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R2x2> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }

    SECTION("integrate on cell list")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.size(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            Array<CellId> cell_list{mesh_1d->numberOfCells() / 2 + mesh_1d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_1d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            Array<CellId> cell_list{mesh_2d->numberOfCells() / 2 + mesh_2d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_2d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            Array<CellId> cell_list{mesh_3d->numberOfCells() / 2 + mesh_3d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_3d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }
  }

  SECTION("Gauss-Lobatto quadrature")
  {
    auto quadrature_descriptor = GaussLobattoQuadratureDescriptor(3);

    SECTION("integrate on all cells")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.numberOfItems(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<double> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R3> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_1d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_1d, cell_integral);

              Array<R2x2> integrate_value(mesh_1d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<double> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R3> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_2d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_2d, cell_integral);

              Array<R2x2> integrate_value(mesh_2d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<double> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<double> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R3> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R3> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R3(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              CellValue<R2x2> cell_integral{mesh_3d->connectivity()};
              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };
              CellIntegrator::integrateTo(f, quadrature_descriptor, *mesh_3d, cell_integral);

              Array<R2x2> integrate_value(mesh_3d->numberOfCells());

              ASTNode::setStackDetails(false);

              IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, integrate_value);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }

    SECTION("integrate on cell list")
    {
      auto same_item_integral = [](auto f, auto g) -> bool {
        using ItemIdType = typename decltype(f)::index_type;
        for (ItemIdType item_id = 0; item_id < f.size(); ++item_id) {
          if (f[item_id] != g[item_id]) {
            return false;
          }
        }

        return true;
      };

      SECTION("1D")
      {
        constexpr size_t Dimension = 1;

        std::array mesh_list = MeshDataBaseForTests::get().all1DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_1d_v = named_mesh.mesh();
            auto mesh_1d   = mesh_1d_v->get<Mesh<1>>();

            Array<CellId> cell_list{mesh_1d->numberOfCells() / 2 + mesh_1d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_1d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_1d: R^1 -> R, x -> 2 * exp(x[0]) + 3;
let R3_1d: R^1 -> R^3, x -> [2 * exp(x[0]) + 3, x[0] - 2, 3];
let R2x2_1d: R^1 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0])], [3, x[0] * x[0]]];
)";
            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 1d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * std::exp(x[0]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 1d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 { return R3{2 * exp(x[0]) + 3, x[0] - 2, 3}; };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_1d, cell_list);
              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 1d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_1d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[0]) + 3, sin(x[0] - 2 * x[0]), 3, x[0] * x[0]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_1d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_1d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("2D")
      {
        constexpr size_t Dimension = 2;

        std::array mesh_list = MeshDataBaseForTests::get().all2DMeshes();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_2d_v = named_mesh.mesh();
            auto mesh_2d   = mesh_2d_v->get<Mesh<2>>();

            Array<CellId> cell_list{mesh_2d->numberOfCells() / 2 + mesh_2d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_2d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*exp(x[0])*sin(x[1])+3;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 2d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 2d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + 3, x[0] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 2d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_2d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3, sin(x[0] - 2 * x[1]), 3, x[0] * x[1]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_2d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_2d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }

      SECTION("3D")
      {
        constexpr size_t Dimension = 3;
        using NamedMesh            = MeshDataBaseForTests::NamedMesh;

        std::vector<NamedMesh> mesh_list = [] {
          std::vector<NamedMesh> extended_mesh_list;
          std::array mesh_array = MeshDataBaseForTests::get().all3DMeshes();
          for (size_t i = 0; i < mesh_array.size(); ++i) {
            extended_mesh_list.push_back(MeshDataBaseForTests::get().all3DMeshes()[i]);
          }
          extended_mesh_list.push_back(NamedMesh("diamond dual", DualMeshManager::instance().getDiamondDualMesh(
                                                                   MeshDataBaseForTests::get().hybrid3DMesh())));
          return extended_mesh_list;
        }();

        for (const auto& named_mesh : mesh_list) {
          SECTION(named_mesh.name())
          {
            auto mesh_3d_v = named_mesh.mesh();
            auto mesh_3d   = mesh_3d_v->get<Mesh<3>>();

            Array<CellId> cell_list{mesh_3d->numberOfCells() / 2 + mesh_3d->numberOfCells() % 2};

            {
              size_t k = 0;
              for (CellId cell_id = 0; cell_id < mesh_3d->numberOfCells(); ++(++cell_id), ++k) {
                cell_list[k] = cell_id;
              }

              REQUIRE(k == cell_list.size());
            }

            std::string_view data = R"(
import math;
let scalar_3d: R^3 -> R, x -> 2 * exp(x[0]) * sin(x[1]) * x[2] + 3;
let R3_3d: R^3 -> R^3, x -> [2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3];
let R2x2_3d: R^3 -> R^2x2, x -> [[2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2])], [3, x[0] * x[1] * x[2]]];
)";

            TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

            auto ast = ASTBuilder::build(input);

            ASTModulesImporter{*ast};
            ASTNodeTypeCleaner<language::import_instruction>{*ast};

            ASTSymbolTableBuilder{*ast};
            ASTNodeDataTypeBuilder{*ast};

            ASTNodeTypeCleaner<language::var_declaration>{*ast};
            ASTNodeTypeCleaner<language::fct_declaration>{*ast};
            ASTNodeExpressionBuilder{*ast};

            std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

            // ensure that variables are declared at this point
            TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

            SECTION("scalar 3d")
            {
              auto [i_symbol, found] = symbol_table->find("scalar_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> double { return 2 * exp(x[0]) * sin(x[1]) * x[2] + 3; };

              Array<const double> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const double> integrate_value =
                IntegrateOnCells<double(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                           *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("vector 3d")
            {
              using R3               = TinyVector<3>;
              auto [i_symbol, found] = symbol_table->find("R3_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R3 {
                return R3{2 * exp(x[0]) * sin(x[1]) + x[2] + 3, x[0] * x[2] - 2 * x[1], 3};
              };

              Array<const R3> cell_integral = CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R3> integrate_value =
                IntegrateOnCells<R3(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                       *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }

            SECTION("matrix 3d")
            {
              using R2x2             = TinyMatrix<2>;
              auto [i_symbol, found] = symbol_table->find("R2x2_3d", position);
              REQUIRE(found);
              REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

              FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

              auto f = [](const TinyVector<Dimension>& x) -> R2x2 {
                return R2x2{2 * exp(x[0]) * sin(x[1]) + 3 * cos(x[2]), sin(x[0] - 2 * x[1] * x[2]), 3,
                            x[0] * x[1] * x[2]};
              };

              Array<const R2x2> cell_integral =
                CellIntegrator::integrate(f, quadrature_descriptor, *mesh_3d, cell_list);

              ASTNode::setStackDetails(false);

              Array<const R2x2> integrate_value =
                IntegrateOnCells<R2x2(TinyVector<Dimension>)>::integrate(function_symbol_id, quadrature_descriptor,
                                                                         *mesh_3d, cell_list);

              ASTNode::setStackDetails(true);

              REQUIRE(same_item_integral(cell_integral, integrate_value));
            }
            ast->m_symbol_table->clearValues();
          }
        }
      }
    }
  }

  SECTION("polynomial 2d mesh (scalar data)")
  {
    constexpr size_t Dimension = 2;

    using Rd = TinyVector<Dimension>;

    auto integrate_on_P2_triangle = [](const Rd& a0, const Rd& a1, const Rd& a2,   //
                                       const Rd& a01, const Rd& a12, const Rd& a20) {
      return (44 * a20[0] * a2[1] * a2[1] * a2[1] + 9 * a1[0] * a2[1] * a2[1] * a2[1] -
              44 * a12[0] * a2[1] * a2[1] * a2[1] - 9 * a0[0] * a2[1] * a2[1] * a2[1] -
              44 * a20[1] * a2[0] * a2[1] * a2[1] - 88 * a20[0] * a2[0] * a2[1] * a2[1] -
              9 * a1[1] * a2[0] * a2[1] * a2[1] - 18 * a1[0] * a2[0] * a2[1] * a2[1] +
              44 * a12[1] * a2[0] * a2[1] * a2[1] + 88 * a12[0] * a2[0] * a2[1] * a2[1] +
              9 * a0[1] * a2[0] * a2[1] * a2[1] + 18 * a0[0] * a2[0] * a2[1] * a2[1] +
              48 * a20[0] * a20[1] * a2[1] * a2[1] - 4 * a0[0] * a20[1] * a2[1] * a2[1] -
              48 * a20[0] * a20[0] * a2[1] * a2[1] - 12 * a0[1] * a20[0] * a2[1] * a2[1] +
              16 * a0[0] * a20[0] * a2[1] * a2[1] + 168 * a20[0] * a2[1] * a2[1] - 3 * a1[0] * a1[1] * a2[1] * a2[1] +
              12 * a12[0] * a1[1] * a2[1] * a2[1] + 3 * a1[0] * a1[0] * a2[1] * a2[1] +
              4 * a12[1] * a1[0] * a2[1] * a2[1] - 16 * a12[0] * a1[0] * a2[1] * a2[1] + 28 * a1[0] * a2[1] * a2[1] -
              48 * a12[0] * a12[1] * a2[1] * a2[1] + 48 * a12[0] * a12[0] * a2[1] * a2[1] -
              168 * a12[0] * a2[1] * a2[1] + 3 * a0[0] * a0[1] * a2[1] * a2[1] - 3 * a0[0] * a0[0] * a2[1] * a2[1] -
              28 * a0[0] * a2[1] * a2[1] + 88 * a20[1] * a2[0] * a2[0] * a2[1] - 44 * a20[0] * a2[0] * a2[0] * a2[1] +
              18 * a1[1] * a2[0] * a2[0] * a2[1] - 9 * a1[0] * a2[0] * a2[0] * a2[1] -
              88 * a12[1] * a2[0] * a2[0] * a2[1] + 44 * a12[0] * a2[0] * a2[0] * a2[1] -
              18 * a0[1] * a2[0] * a2[0] * a2[1] + 9 * a0[0] * a2[0] * a2[0] * a2[1] -
              48 * a20[1] * a20[1] * a2[0] * a2[1] + 16 * a0[1] * a20[1] * a2[0] * a2[1] -
              8 * a0[0] * a20[1] * a2[0] * a2[1] - 168 * a20[1] * a2[0] * a2[1] - 48 * a20[0] * a20[0] * a2[0] * a2[1] +
              8 * a0[1] * a20[0] * a2[0] * a2[1] + 16 * a0[0] * a20[0] * a2[0] * a2[1] - 252 * a20[0] * a2[0] * a2[1] +
              3 * a1[1] * a1[1] * a2[0] * a2[1] - 16 * a12[1] * a1[1] * a2[0] * a2[1] -
              8 * a12[0] * a1[1] * a2[0] * a2[1] - 28 * a1[1] * a2[0] * a2[1] + 3 * a1[0] * a1[0] * a2[0] * a2[1] +
              8 * a12[1] * a1[0] * a2[0] * a2[1] - 16 * a12[0] * a1[0] * a2[0] * a2[1] - 42 * a1[0] * a2[0] * a2[1] +
              48 * a12[1] * a12[1] * a2[0] * a2[1] + 168 * a12[1] * a2[0] * a2[1] +
              48 * a12[0] * a12[0] * a2[0] * a2[1] + 252 * a12[0] * a2[0] * a2[1] - 3 * a0[1] * a0[1] * a2[0] * a2[1] +
              28 * a0[1] * a2[0] * a2[1] - 3 * a0[0] * a0[0] * a2[0] * a2[1] + 42 * a0[0] * a2[0] * a2[1] +
              64 * a20[0] * a20[1] * a20[1] * a2[1] - 16 * a0[0] * a20[1] * a20[1] * a2[1] -
              128 * a20[0] * a20[0] * a20[1] * a2[1] + 32 * a0[0] * a20[0] * a20[1] * a2[1] +
              224 * a20[0] * a20[1] * a2[1] - 16 * a0[0] * a0[1] * a20[1] * a2[1] +
              16 * a0[0] * a0[0] * a20[1] * a2[1] - 56 * a0[0] * a20[1] * a2[1] -
              64 * a20[0] * a20[0] * a20[0] * a2[1] + 16 * a0[0] * a20[0] * a20[0] * a2[1] -
              336 * a20[0] * a20[0] * a2[1] + 12 * a0[1] * a0[1] * a20[0] * a2[1] - 8 * a0[0] * a0[1] * a20[0] * a2[1] +
              4 * a0[0] * a0[0] * a20[0] * a2[1] + 84 * a0[0] * a20[0] * a2[1] - 560 * a20[0] * a2[1] +
              9 * a1[0] * a1[1] * a1[1] * a2[1] - 12 * a12[0] * a1[1] * a1[1] * a2[1] -
              18 * a1[0] * a1[0] * a1[1] * a2[1] + 16 * a12[1] * a1[0] * a1[1] * a2[1] +
              8 * a12[0] * a1[0] * a1[1] * a2[1] + 28 * a1[0] * a1[1] * a2[1] - 9 * a1[0] * a1[0] * a1[0] * a2[1] -
              16 * a12[1] * a1[0] * a1[0] * a2[1] - 4 * a12[0] * a1[0] * a1[0] * a2[1] - 42 * a1[0] * a1[0] * a2[1] +
              16 * a12[1] * a12[1] * a1[0] * a2[1] - 32 * a12[0] * a12[1] * a1[0] * a2[1] +
              56 * a12[1] * a1[0] * a2[1] - 16 * a12[0] * a12[0] * a1[0] * a2[1] - 84 * a12[0] * a1[0] * a2[1] -
              140 * a1[0] * a2[1] - 64 * a12[0] * a12[1] * a12[1] * a2[1] + 128 * a12[0] * a12[0] * a12[1] * a2[1] -
              224 * a12[0] * a12[1] * a2[1] + 64 * a12[0] * a12[0] * a12[0] * a2[1] + 336 * a12[0] * a12[0] * a2[1] +
              560 * a12[0] * a2[1] - 9 * a0[0] * a0[1] * a0[1] * a2[1] + 18 * a0[0] * a0[0] * a0[1] * a2[1] -
              28 * a0[0] * a0[1] * a2[1] + 9 * a0[0] * a0[0] * a0[0] * a2[1] + 42 * a0[0] * a0[0] * a2[1] +
              140 * a0[0] * a2[1] + 44 * a20[1] * a2[0] * a2[0] * a2[0] + 9 * a1[1] * a2[0] * a2[0] * a2[0] -
              44 * a12[1] * a2[0] * a2[0] * a2[0] - 9 * a0[1] * a2[0] * a2[0] * a2[0] +
              48 * a20[1] * a20[1] * a2[0] * a2[0] + 48 * a20[0] * a20[1] * a2[0] * a2[0] -
              16 * a0[1] * a20[1] * a2[0] * a2[0] - 12 * a0[0] * a20[1] * a2[0] * a2[0] + 252 * a20[1] * a2[0] * a2[0] -
              4 * a0[1] * a20[0] * a2[0] * a2[0] - 3 * a1[1] * a1[1] * a2[0] * a2[0] -
              3 * a1[0] * a1[1] * a2[0] * a2[0] + 16 * a12[1] * a1[1] * a2[0] * a2[0] +
              4 * a12[0] * a1[1] * a2[0] * a2[0] + 42 * a1[1] * a2[0] * a2[0] + 12 * a12[1] * a1[0] * a2[0] * a2[0] -
              48 * a12[1] * a12[1] * a2[0] * a2[0] - 48 * a12[0] * a12[1] * a2[0] * a2[0] -
              252 * a12[1] * a2[0] * a2[0] + 3 * a0[1] * a0[1] * a2[0] * a2[0] + 3 * a0[0] * a0[1] * a2[0] * a2[0] -
              42 * a0[1] * a2[0] * a2[0] - 64 * a20[1] * a20[1] * a20[1] * a2[0] +
              128 * a20[0] * a20[1] * a20[1] * a2[0] + 16 * a0[1] * a20[1] * a20[1] * a2[0] -
              224 * a20[1] * a20[1] * a2[0] + 64 * a20[0] * a20[0] * a20[1] * a2[0] -
              32 * a0[1] * a20[0] * a20[1] * a2[0] + 336 * a20[0] * a20[1] * a2[0] +
              4 * a0[1] * a0[1] * a20[1] * a2[0] + 8 * a0[0] * a0[1] * a20[1] * a2[0] + 56 * a0[1] * a20[1] * a2[0] +
              12 * a0[0] * a0[0] * a20[1] * a2[0] + 560 * a20[1] * a2[0] - 16 * a0[1] * a20[0] * a20[0] * a2[0] -
              16 * a0[1] * a0[1] * a20[0] * a2[0] - 16 * a0[0] * a0[1] * a20[0] * a2[0] - 84 * a0[1] * a20[0] * a2[0] -
              9 * a1[1] * a1[1] * a1[1] * a2[0] + 18 * a1[0] * a1[1] * a1[1] * a2[0] -
              4 * a12[1] * a1[1] * a1[1] * a2[0] + 16 * a12[0] * a1[1] * a1[1] * a2[0] - 28 * a1[1] * a1[1] * a2[0] +
              9 * a1[0] * a1[0] * a1[1] * a2[0] - 8 * a12[1] * a1[0] * a1[1] * a2[0] +
              16 * a12[0] * a1[0] * a1[1] * a2[0] + 42 * a1[0] * a1[1] * a2[0] - 16 * a12[1] * a12[1] * a1[1] * a2[0] +
              32 * a12[0] * a12[1] * a1[1] * a2[0] - 56 * a12[1] * a1[1] * a2[0] +
              16 * a12[0] * a12[0] * a1[1] * a2[0] + 84 * a12[0] * a1[1] * a2[0] + 140 * a1[1] * a2[0] -
              12 * a12[1] * a1[0] * a1[0] * a2[0] + 64 * a12[1] * a12[1] * a12[1] * a2[0] -
              128 * a12[0] * a12[1] * a12[1] * a2[0] + 224 * a12[1] * a12[1] * a2[0] -
              64 * a12[0] * a12[0] * a12[1] * a2[0] - 336 * a12[0] * a12[1] * a2[0] - 560 * a12[1] * a2[0] +
              9 * a0[1] * a0[1] * a0[1] * a2[0] - 18 * a0[0] * a0[1] * a0[1] * a2[0] + 28 * a0[1] * a0[1] * a2[0] -
              9 * a0[0] * a0[0] * a0[1] * a2[0] - 42 * a0[0] * a0[1] * a2[0] - 140 * a0[1] * a2[0] +
              64 * a0[0] * a20[1] * a20[1] * a20[1] - 64 * a0[1] * a20[0] * a20[1] * a20[1] -
              128 * a0[0] * a20[0] * a20[1] * a20[1] + 48 * a0[0] * a0[1] * a20[1] * a20[1] -
              48 * a0[0] * a0[0] * a20[1] * a20[1] + 224 * a0[0] * a20[1] * a20[1] +
              128 * a0[1] * a20[0] * a20[0] * a20[1] - 64 * a0[0] * a20[0] * a20[0] * a20[1] -
              48 * a0[1] * a0[1] * a20[0] * a20[1] - 224 * a0[1] * a20[0] * a20[1] -
              48 * a0[0] * a0[0] * a20[0] * a20[1] - 336 * a0[0] * a20[0] * a20[1] +
              44 * a0[0] * a0[1] * a0[1] * a20[1] - 88 * a0[0] * a0[0] * a0[1] * a20[1] + 168 * a0[0] * a0[1] * a20[1] -
              44 * a0[0] * a0[0] * a0[0] * a20[1] - 252 * a0[0] * a0[0] * a20[1] - 560 * a0[0] * a20[1] +
              64 * a0[1] * a20[0] * a20[0] * a20[0] + 48 * a0[1] * a0[1] * a20[0] * a20[0] +
              48 * a0[0] * a0[1] * a20[0] * a20[0] + 336 * a0[1] * a20[0] * a20[0] -
              44 * a0[1] * a0[1] * a0[1] * a20[0] + 88 * a0[0] * a0[1] * a0[1] * a20[0] - 168 * a0[1] * a0[1] * a20[0] +
              44 * a0[0] * a0[0] * a0[1] * a20[0] + 252 * a0[0] * a0[1] * a20[0] + 560 * a0[1] * a20[0] +
              44 * a12[0] * a1[1] * a1[1] * a1[1] + 9 * a0[0] * a1[1] * a1[1] * a1[1] -
              44 * a01[0] * a1[1] * a1[1] * a1[1] - 44 * a12[1] * a1[0] * a1[1] * a1[1] -
              88 * a12[0] * a1[0] * a1[1] * a1[1] - 9 * a0[1] * a1[0] * a1[1] * a1[1] -
              18 * a0[0] * a1[0] * a1[1] * a1[1] + 44 * a01[1] * a1[0] * a1[1] * a1[1] +
              88 * a01[0] * a1[0] * a1[1] * a1[1] + 48 * a12[0] * a12[1] * a1[1] * a1[1] -
              48 * a12[0] * a12[0] * a1[1] * a1[1] + 168 * a12[0] * a1[1] * a1[1] - 3 * a0[0] * a0[1] * a1[1] * a1[1] +
              12 * a01[0] * a0[1] * a1[1] * a1[1] + 3 * a0[0] * a0[0] * a1[1] * a1[1] +
              4 * a01[1] * a0[0] * a1[1] * a1[1] - 16 * a01[0] * a0[0] * a1[1] * a1[1] + 28 * a0[0] * a1[1] * a1[1] -
              48 * a01[0] * a01[1] * a1[1] * a1[1] + 48 * a01[0] * a01[0] * a1[1] * a1[1] -
              168 * a01[0] * a1[1] * a1[1] + 88 * a12[1] * a1[0] * a1[0] * a1[1] - 44 * a12[0] * a1[0] * a1[0] * a1[1] +
              18 * a0[1] * a1[0] * a1[0] * a1[1] - 9 * a0[0] * a1[0] * a1[0] * a1[1] -
              88 * a01[1] * a1[0] * a1[0] * a1[1] + 44 * a01[0] * a1[0] * a1[0] * a1[1] -
              48 * a12[1] * a12[1] * a1[0] * a1[1] - 168 * a12[1] * a1[0] * a1[1] -
              48 * a12[0] * a12[0] * a1[0] * a1[1] - 252 * a12[0] * a1[0] * a1[1] + 3 * a0[1] * a0[1] * a1[0] * a1[1] -
              16 * a01[1] * a0[1] * a1[0] * a1[1] - 8 * a01[0] * a0[1] * a1[0] * a1[1] - 28 * a0[1] * a1[0] * a1[1] +
              3 * a0[0] * a0[0] * a1[0] * a1[1] + 8 * a01[1] * a0[0] * a1[0] * a1[1] -
              16 * a01[0] * a0[0] * a1[0] * a1[1] - 42 * a0[0] * a1[0] * a1[1] + 48 * a01[1] * a01[1] * a1[0] * a1[1] +
              168 * a01[1] * a1[0] * a1[1] + 48 * a01[0] * a01[0] * a1[0] * a1[1] + 252 * a01[0] * a1[0] * a1[1] +
              64 * a12[0] * a12[1] * a12[1] * a1[1] - 128 * a12[0] * a12[0] * a12[1] * a1[1] +
              224 * a12[0] * a12[1] * a1[1] - 64 * a12[0] * a12[0] * a12[0] * a1[1] - 336 * a12[0] * a12[0] * a1[1] -
              560 * a12[0] * a1[1] + 9 * a0[0] * a0[1] * a0[1] * a1[1] - 12 * a01[0] * a0[1] * a0[1] * a1[1] -
              18 * a0[0] * a0[0] * a0[1] * a1[1] + 16 * a01[1] * a0[0] * a0[1] * a1[1] +
              8 * a01[0] * a0[0] * a0[1] * a1[1] + 28 * a0[0] * a0[1] * a1[1] - 9 * a0[0] * a0[0] * a0[0] * a1[1] -
              16 * a01[1] * a0[0] * a0[0] * a1[1] - 4 * a01[0] * a0[0] * a0[0] * a1[1] - 42 * a0[0] * a0[0] * a1[1] +
              16 * a01[1] * a01[1] * a0[0] * a1[1] - 32 * a01[0] * a01[1] * a0[0] * a1[1] +
              56 * a01[1] * a0[0] * a1[1] - 16 * a01[0] * a01[0] * a0[0] * a1[1] - 84 * a01[0] * a0[0] * a1[1] -
              140 * a0[0] * a1[1] - 64 * a01[0] * a01[1] * a01[1] * a1[1] + 128 * a01[0] * a01[0] * a01[1] * a1[1] -
              224 * a01[0] * a01[1] * a1[1] + 64 * a01[0] * a01[0] * a01[0] * a1[1] + 336 * a01[0] * a01[0] * a1[1] +
              560 * a01[0] * a1[1] + 44 * a12[1] * a1[0] * a1[0] * a1[0] + 9 * a0[1] * a1[0] * a1[0] * a1[0] -
              44 * a01[1] * a1[0] * a1[0] * a1[0] + 48 * a12[1] * a12[1] * a1[0] * a1[0] +
              48 * a12[0] * a12[1] * a1[0] * a1[0] + 252 * a12[1] * a1[0] * a1[0] - 3 * a0[1] * a0[1] * a1[0] * a1[0] -
              3 * a0[0] * a0[1] * a1[0] * a1[0] + 16 * a01[1] * a0[1] * a1[0] * a1[0] +
              4 * a01[0] * a0[1] * a1[0] * a1[0] + 42 * a0[1] * a1[0] * a1[0] + 12 * a01[1] * a0[0] * a1[0] * a1[0] -
              48 * a01[1] * a01[1] * a1[0] * a1[0] - 48 * a01[0] * a01[1] * a1[0] * a1[0] -
              252 * a01[1] * a1[0] * a1[0] - 64 * a12[1] * a12[1] * a12[1] * a1[0] +
              128 * a12[0] * a12[1] * a12[1] * a1[0] - 224 * a12[1] * a12[1] * a1[0] +
              64 * a12[0] * a12[0] * a12[1] * a1[0] + 336 * a12[0] * a12[1] * a1[0] + 560 * a12[1] * a1[0] -
              9 * a0[1] * a0[1] * a0[1] * a1[0] + 18 * a0[0] * a0[1] * a0[1] * a1[0] -
              4 * a01[1] * a0[1] * a0[1] * a1[0] + 16 * a01[0] * a0[1] * a0[1] * a1[0] - 28 * a0[1] * a0[1] * a1[0] +
              9 * a0[0] * a0[0] * a0[1] * a1[0] - 8 * a01[1] * a0[0] * a0[1] * a1[0] +
              16 * a01[0] * a0[0] * a0[1] * a1[0] + 42 * a0[0] * a0[1] * a1[0] - 16 * a01[1] * a01[1] * a0[1] * a1[0] +
              32 * a01[0] * a01[1] * a0[1] * a1[0] - 56 * a01[1] * a0[1] * a1[0] +
              16 * a01[0] * a01[0] * a0[1] * a1[0] + 84 * a01[0] * a0[1] * a1[0] + 140 * a0[1] * a1[0] -
              12 * a01[1] * a0[0] * a0[0] * a1[0] + 64 * a01[1] * a01[1] * a01[1] * a1[0] -
              128 * a01[0] * a01[1] * a01[1] * a1[0] + 224 * a01[1] * a01[1] * a1[0] -
              64 * a01[0] * a01[0] * a01[1] * a1[0] - 336 * a01[0] * a01[1] * a1[0] - 560 * a01[1] * a1[0] +
              44 * a01[0] * a0[1] * a0[1] * a0[1] - 44 * a01[1] * a0[0] * a0[1] * a0[1] -
              88 * a01[0] * a0[0] * a0[1] * a0[1] + 48 * a01[0] * a01[1] * a0[1] * a0[1] -
              48 * a01[0] * a01[0] * a0[1] * a0[1] + 168 * a01[0] * a0[1] * a0[1] +
              88 * a01[1] * a0[0] * a0[0] * a0[1] - 44 * a01[0] * a0[0] * a0[0] * a0[1] -
              48 * a01[1] * a01[1] * a0[0] * a0[1] - 168 * a01[1] * a0[0] * a0[1] -
              48 * a01[0] * a01[0] * a0[0] * a0[1] - 252 * a01[0] * a0[0] * a0[1] +
              64 * a01[0] * a01[1] * a01[1] * a0[1] - 128 * a01[0] * a01[0] * a01[1] * a0[1] +
              224 * a01[0] * a01[1] * a0[1] - 64 * a01[0] * a01[0] * a01[0] * a0[1] - 336 * a01[0] * a01[0] * a0[1] -
              560 * a01[0] * a0[1] + 44 * a01[1] * a0[0] * a0[0] * a0[0] + 48 * a01[1] * a01[1] * a0[0] * a0[0] +
              48 * a01[0] * a01[1] * a0[0] * a0[0] + 252 * a01[1] * a0[0] * a0[0] -
              64 * a01[1] * a01[1] * a01[1] * a0[0] + 128 * a01[0] * a01[1] * a01[1] * a0[0] -
              224 * a01[1] * a01[1] * a0[0] + 64 * a01[0] * a01[0] * a01[1] * a0[0] + 336 * a01[0] * a01[1] * a0[0] +
              560 * a01[1] * a0[0]) /
             420;
    };

    auto L1_error = [=](auto& p_mesh, auto& cell_value) -> double {
      using MeshType = mesh_type_t<decltype(p_mesh)>;
      static_assert(is_polynomial_mesh_v<MeshType>);

      auto cell_type           = p_mesh->connectivity().cellType();
      auto cell_to_face_matrix = p_mesh->connectivity().cellToFaceMatrix();
      auto cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();

      const auto xr = p_mesh->xr();
      const auto xl = p_mesh->xl();
      const auto xj = MeshDataManager::instance().getMeshData(*p_mesh).xj();

      double error = 0;
      for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
        const auto cell_to_face = cell_to_face_matrix[cell_id];
        const auto cell_to_node = cell_to_node_matrix[cell_id];
        switch (cell_type[cell_id]) {
        case CellType::Triangle: {
          const Rd& a0 = xr[cell_to_node[0]];
          const Rd& a1 = xr[cell_to_node[1]];
          const Rd& a2 = xr[cell_to_node[2]];

          const Rd& a01 = xl[cell_to_face[2]][0];
          const Rd& a12 = xl[cell_to_face[0]][0];
          const Rd& a20 = xl[cell_to_face[1]][0];

          error += std::abs(cell_value[cell_id] - integrate_on_P2_triangle(a0, a1, a2,   //
                                                                           a01, a12, a20));
          break;
        }
        case CellType::Quadrangle: {
          const Rd& a00 = xr[cell_to_node[0]];
          const Rd& a01 = xl[cell_to_face[3]][0];
          const Rd& a02 = xr[cell_to_node[3]];

          const Rd& a10 = xl[cell_to_face[0]][0];
          const Rd& a11 = xj[cell_id];
          const Rd& a20 = xr[cell_to_node[1]];

          const Rd& a21 = xl[cell_to_face[1]][0];
          const Rd& a12 = xl[cell_to_face[2]][0];
          const Rd& a22 = xr[cell_to_node[2]];

          const double int_p =                          //
            integrate_on_P2_triangle(a00, a22, a02,     //
                                     a11, a12, a01) +   //
            integrate_on_P2_triangle(a00, a20, a22,     //
                                     a10, a21, a11);

          error += std::abs(cell_value[cell_id] - int_p);
          break;
        }
        default: {
          throw UnexpectedError("unexpected cell type");
        }
        }
      }

      return error;
    };

    auto polygonal_mesh_2d_v = MeshDataBaseForTests::get().hybrid2DMesh();

    auto mesh_2d_v        = PolynomialMeshBuilder{polygonal_mesh_2d_v, 2}.mesh();
    auto straight_mesh_2d = mesh_2d_v->get<PolynomialMesh<2>>();

    const double pi = std::acos(-1);
    const double L  = 2 / pi;

    auto xr = copy(straight_mesh_2d->xr());
    auto xl = copy(straight_mesh_2d->xl());
    auto T  = [=](const Rd& x) {
      return Rd{x[0] + 0.5 * std::cos(x[0] / L) * std::sin(x[1] / L),
                x[1] - 0.5 * std::sin(x[0] / L) * std::cos(x[1] / L)};
    };

    for (NodeId node_id = 0; node_id < straight_mesh_2d->numberOfNodes(); ++node_id) {
      xr[node_id] = T(xr[node_id]);
    }
    for (FaceId face_id = 0; face_id < straight_mesh_2d->numberOfNodes(); ++face_id) {
      for (size_t i = 0; i < xl.sizeOfArrays(); ++i) {
        xl[face_id][i] = T(xl[face_id][i]);
      }
    }

    auto mesh_2d = std::make_shared<const PolynomialMesh<2>>(straight_mesh_2d->shared_connectivity(), xr, xl);

    std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x ->  2 + 3 * x[0] - 2 * x[1] + x[0] * x[0] + 2 * x[0] * x[1] - x[1] * x[1];
)";

    TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

    auto ast = ASTBuilder::build(input);

    ASTModulesImporter{*ast};
    ASTNodeTypeCleaner<language::import_instruction>{*ast};

    ASTSymbolTableBuilder{*ast};
    ASTNodeDataTypeBuilder{*ast};

    ASTNodeTypeCleaner<language::var_declaration>{*ast};
    ASTNodeTypeCleaner<language::fct_declaration>{*ast};
    ASTNodeExpressionBuilder{*ast};

    std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

    // ensure that variables are declared at this point
    TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

    auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
    REQUIRE(found);
    REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

    FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

    // polynomial is degree 2, transformation is degree 2 and detJ is
    // degree 3 => 7
    auto quadrature_descriptor = GaussLegendreQuadratureDescriptor(7);

    ASTNode::setStackDetails(false);

    CellValue<double> integrate_value(mesh_2d->connectivity());
    IntegrateOnCells<double(TinyVector<Dimension>)>::integrateTo(function_symbol_id, quadrature_descriptor, *mesh_2d,
                                                                 integrate_value);

    ASTNode::setStackDetails(true);

    REQUIRE(L1_error(mesh_2d, integrate_value) == Catch::Approx(0.).margin(1E-13));
  }
}
