#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineCubicTransformation.hpp>
#include <geometry/LineParabolicTransformation.hpp>
#include <geometry/LineTransformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("LineCubicTransformation", "[geometry]")
{
  SECTION("2D")
  {
    SECTION("b and c are at thirds")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 d{2.7, 0.7};
      const R2 b = a + 1. / 3 * (d - a);
      const R2 c = a + 2. / 3 * (d - a);

      const LineCubicTransformation<2> t(a, b, c, d);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{-1. / 3})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{-1. / 3})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1. / 3})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1. / 3})[1] == Catch::Approx(c[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(d[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(d[1]));

      REQUIRE(l2Norm(t.velocity(R1{-1}) - 0.5 * (d - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - 0.5 * (d - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - 0.5 * (d - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - 0.5 * (d - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - 0.5 * (d - a)) == Catch::Approx(0).margin(1E-14));

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(l2Norm(0.5 * (d - a))));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(l2Norm(0.5 * (d - a))));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(l2Norm(0.5 * (d - a))));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(l2Norm(0.5 * (d - a))));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(l2Norm(0.5 * (d - a))));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(2));

      double sum = 0;
      for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
        sum += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
      }

      REQUIRE(sum == Catch::Approx(24.8585155630822));
    }

    SECTION("b,c are aligned")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 d{2.7, 0.7};
      const R2 b = 0.3 * a + 0.7 * d;
      const R2 c = 0.2 * a + 0.8 * d;

      const LineCubicTransformation<2> t(a, b, c, d);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{-1. / 3})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{-1. / 3})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1. / 3})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1. / 3})[1] == Catch::Approx(c[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(d[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(d[1]));

      auto v = [&](double s_hat) {
        return                                                        //
          -9. / 16 * (3 * s_hat * s_hat - 2 * s_hat - 1. / 9) * a     //
          + 27. / 16 * (3 * s_hat * s_hat - 2. / 3 * s_hat - 1) * b   //
          - 27. / 16 * (3 * s_hat * s_hat + 2. / 3 * s_hat - 1) * c   //
          + 9. / 16 * (3 * s_hat * s_hat + 2 * s_hat - 1. / 9) * d;
      };

      REQUIRE(l2Norm(t.velocity(R1{-1}) - v(-1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - v(0)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - v(1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - v(-0.6)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - v(0.37)) == Catch::Approx(0).margin(1e-14));

      auto norm_v = [&](double s_hat) { return l2Norm(v(s_hat)); };

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(norm_v(-1)));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(norm_v(0)));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(norm_v(1)));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(norm_v(-0.6)));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(norm_v(0.37)));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(8));

      double sum = 0;
      for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
        sum += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
      }

      REQUIRE(sum == Catch::Approx(24.8585155630822));
    }

    SECTION("b is not aligned")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 d{2.7, 1.1};
      const R2 b{1.2, 2};
      const R2 c{2, 1.3};

      const LineCubicTransformation<2> t(a, b, c, d);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{-1. / 3})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{-1. / 3})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1. / 3})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1. / 3})[1] == Catch::Approx(c[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(d[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(d[1]));

      auto v = [&](double s_hat) {
        return                                                        //
          -9. / 16 * (3 * s_hat * s_hat - 2 * s_hat - 1. / 9) * a     //
          + 27. / 16 * (3 * s_hat * s_hat - 2. / 3 * s_hat - 1) * b   //
          - 27. / 16 * (3 * s_hat * s_hat + 2. / 3 * s_hat - 1) * c   //
          + 9. / 16 * (3 * s_hat * s_hat + 2 * s_hat - 1. / 9) * d;
      };

      REQUIRE(l2Norm(t.velocity(R1{-1}) - v(-1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - v(0)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - v(1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - v(-0.6)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - v(0.37)) == Catch::Approx(0).margin(1e-14));

      auto norm_v = [&](double s_hat) { return l2Norm(v(s_hat)); };

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(norm_v(-1)));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(norm_v(0)));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(norm_v(1)));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(norm_v(-0.6)));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(norm_v(0.37)));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      double sum_p3_1_edge = 0;
      {
        QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(23));

        for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
          sum_p3_1_edge += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
        }
      }

      double sum_p2 = 0;
      double sum_p1 = 0;
      {
        QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(8));

        const size_t N = 10000;
        for (size_t i_h = 0; i_h < N; ++i_h) {
          const R1 x_hat0 = R1{-1 + i_h * 2. / N};
          const R1 x_hat2 = R1{-1 + (i_h + 1) * 2. / N};
          const R1 x_hat1 = 0.5 * (x_hat0 + x_hat2);

          const R2 a0 = t(x_hat0);
          const R2 a1 = t(x_hat1);
          const R2 a2 = t(x_hat2);

          LineParabolicTransformation<2> t_p2{a0, a1, a2};
          LineTransformation<2> t_p1{a0, a2};

          for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
            sum_p2 += qf.weight(i) * t_p2.velocityNorm(qf.point(i)) * p(t_p2(qf.point(i)));
            sum_p1 += qf.weight(i) * t_p1.velocityNorm() * p(t_p1(qf.point(i)));
          }
        }
      }

      double sum_p3_4_edges = 0;
      {
        QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(23));
        const size_t N          = 4;
        for (size_t i_h = 0; i_h < N; ++i_h) {
          const R1 x_hat0 = R1{-1 + i_h * 2. / N};
          const R1 x_hat3 = R1{-1 + (i_h + 1) * 2. / N};
          const R1 x_hat1 = 2. / 3 * x_hat0 + 1. / 3 * x_hat3;
          const R1 x_hat2 = 1. / 3 * x_hat0 + 2. / 3 * x_hat3;

          const R2 a0 = t(x_hat0);
          const R2 a1 = t(x_hat1);
          const R2 a2 = t(x_hat2);
          const R2 a3 = t(x_hat3);

          const LineCubicTransformation<2> t_p3(a0, a1, a2, a3);

          for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
            sum_p3_4_edges += qf.weight(i) * t_p3.velocityNorm(qf.point(i)) * p(t_p3(qf.point(i)));
          }
        }
      }

      CHECK(sum_p3_1_edge == Catch::Approx(sum_p2).epsilon(1E-3));
      CHECK(sum_p2 == Catch::Approx(sum_p1).epsilon(1E-6));
      CHECK(sum_p3_4_edges == Catch::Approx(sum_p1).epsilon(1E-6));
    }
  }
}
