#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/LineParabolicTransformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("LineParabolicTransformation", "[geometry]")
{
  SECTION("2D")
  {
    SECTION("b is middle")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 c{2.7, 0.7};
      const R2 b = 0.5 * (a + c);

      const LineParabolicTransformation<2> t(a, b, c);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{0})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{0})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(c[1]));

      REQUIRE(l2Norm(t.velocity(R1{-1}) - 0.5 * (c - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - 0.5 * (c - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - 0.5 * (c - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - 0.5 * (c - a)) == Catch::Approx(0).margin(1E-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - 0.5 * (c - a)) == Catch::Approx(0).margin(1E-14));

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(l2Norm(0.5 * (c - a))));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(l2Norm(0.5 * (c - a))));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(l2Norm(0.5 * (c - a))));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(l2Norm(0.5 * (c - a))));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(l2Norm(0.5 * (c - a))));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(2));

      double sum = 0;
      for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
        sum += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
      }

      REQUIRE(sum == Catch::Approx(24.8585155630822));
    }

    SECTION("b is aligned")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 c{2.7, 0.7};
      const R2 b = 0.2 * a + 0.8 * c;

      const LineParabolicTransformation<2> t(a, b, c);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{0})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{0})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(c[1]));

      auto v = [&](double s_hat) { return (s_hat - 0.5) * a - 2 * s_hat * b + (s_hat + 0.5) * c; };

      REQUIRE(l2Norm(t.velocity(R1{-1}) - v(-1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - v(0)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - v(1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - v(-0.6)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - v(0.37)) == Catch::Approx(0).margin(1e-14));

      auto norme_v = [&](double s_hat) { return l2Norm(v(s_hat)); };

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(norme_v(-1)));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(norme_v(0)));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(norme_v(1)));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(norme_v(-0.6)));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(norme_v(0.37)));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(4));

      double sum = 0;
      for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
        sum += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
      }

      REQUIRE(sum == Catch::Approx(24.8585155630822));
    }

    SECTION("b is not aligned")
    {
      using R1 = TinyVector<1>;
      using R2 = TinyVector<2>;

      const R2 a{0.3, 1.2};
      const R2 c{2.7, 0.7};
      const R2 b{1.2, 2};

      const LineParabolicTransformation<2> t(a, b, c);

      REQUIRE(t(R1{-1})[0] == Catch::Approx(a[0]));
      REQUIRE(t(R1{-1})[1] == Catch::Approx(a[1]));

      REQUIRE(t(R1{0})[0] == Catch::Approx(b[0]));
      REQUIRE(t(R1{0})[1] == Catch::Approx(b[1]));

      REQUIRE(t(R1{1})[0] == Catch::Approx(c[0]));
      REQUIRE(t(R1{1})[1] == Catch::Approx(c[1]));

      auto v = [&](double s_hat) { return (s_hat - 0.5) * a - 2 * s_hat * b + (s_hat + 0.5) * c; };

      REQUIRE(l2Norm(t.velocity(R1{-1}) - v(-1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0}) - v(0)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{1}) - v(1)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{-0.6}) - v(-0.6)) == Catch::Approx(0).margin(1e-14));
      REQUIRE(l2Norm(t.velocity(R1{0.37}) - v(0.37)) == Catch::Approx(0).margin(1e-14));

      auto norm_v = [&](double s_hat) { return l2Norm(v(s_hat)); };

      REQUIRE(t.velocityNorm(R1{-1}) == Catch::Approx(norm_v(-1)));
      REQUIRE(t.velocityNorm(R1{0}) == Catch::Approx(norm_v(0)));
      REQUIRE(t.velocityNorm(R1{1}) == Catch::Approx(norm_v(1)));
      REQUIRE(t.velocityNorm(R1{-0.6}) == Catch::Approx(norm_v(-0.6)));
      REQUIRE(t.velocityNorm(R1{0.37}) == Catch::Approx(norm_v(0.37)));

      auto p = [](const R2& X) {
        const double x = X[0];
        const double y = X[1];
        return 2 * x * x + 3 * x - 3 * y * y + y + 2;
      };

      QuadratureFormula<1> qf = QuadratureManager::instance().getLineFormula(GaussQuadratureDescriptor(23));

      double sum = 0;
      for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
        sum += qf.weight(i) * t.velocityNorm(qf.point(i)) * p(t(qf.point(i)));
      }

      REQUIRE(sum == Catch::Approx(19.577181227595).epsilon(1E-6));
    }
  }
}
