#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <Kokkos_Core.hpp>

#include <utils/PugsAssert.hpp>
#include <utils/Types.hpp>

#include <algebra/SmallMatrix.hpp>
#include <algebra/SmallVector.hpp>
#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/GaussLobattoQuadratureDescriptor.hpp>
#include <analysis/GaussQuadratureDescriptor.hpp>
#include <language/ast/ASTBuilder.hpp>
#include <language/ast/ASTModulesImporter.hpp>
#include <language/ast/ASTNodeDataTypeBuilder.hpp>
#include <language/ast/ASTNodeExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionEvaluationExpressionBuilder.hpp>
#include <language/ast/ASTNodeFunctionExpressionBuilder.hpp>
#include <language/ast/ASTNodeTypeCleaner.hpp>
#include <language/ast/ASTSymbolTableBuilder.hpp>
#include <language/utils/IntegrateOnCells.hpp>
#include <language/utils/PugsFunctionAdapter.hpp>
#include <language/utils/SymbolTable.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshDataManager.hpp>
#include <mesh/PolynomialMeshBuilder.hpp>
#include <scheme/DiscreteFunctionDPk.hpp>
#include <scheme/DiscreteFunctionDPkVariant.hpp>
#include <scheme/DiscreteFunctionP0.hpp>
#include <scheme/DiscreteFunctionVariant.hpp>
#include <scheme/PolynomialReconstruction.hpp>

#include <MeshDataBaseForTests.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("PolynomialP2Reconstruction", "[scheme]")
{
  SECTION("degree 2")
  {
    auto build_parabolic_mesh =
      [](const std::shared_ptr<const MeshVariant>& polygonal_mesh_2d_v) -> std::shared_ptr<const MeshVariant> {
      using Rd = TinyVector<2>;

      auto mesh_2d_v        = PolynomialMeshBuilder{polygonal_mesh_2d_v, 2}.mesh();
      auto straight_mesh_2d = mesh_2d_v->get<PolynomialMesh<2>>();

      const double pi = std::acos(-1);
      const double L  = 2 / pi;

      auto xr = copy(straight_mesh_2d->xr());
      auto xl = copy(straight_mesh_2d->xl());
      auto T  = [=](const Rd& x) {
        return Rd{x[0] + 0.5 * std::cos(x[0] / L) * std::sin(x[1] / L),
                  x[1] - 0.5 * std::sin(x[0] / L) * std::cos(x[1] / L)};
      };

      for (NodeId node_id = 0; node_id < straight_mesh_2d->numberOfNodes(); ++node_id) {
        xr[node_id] = T(xr[node_id]);
      }
      for (FaceId face_id = 0; face_id < straight_mesh_2d->numberOfNodes(); ++face_id) {
        for (size_t i = 0; i < xl.sizeOfArrays(); ++i) {
          xl[face_id][i] = T(xl[face_id][i]);
        }
      }

      auto mesh_2d = std::make_shared<const PolynomialMesh<2>>(straight_mesh_2d->shared_connectivity(), xr, xl);
      return std::make_shared<const MeshVariant>(mesh_2d);
    };

    using NamedMesh = MeshDataBaseForTests::NamedMesh;

    auto mesh_list =
      std::array{NamedMesh{"cartesian 2d mesh", MeshDataBaseForTests::get().cartesian2DMesh()},
                 NamedMesh{"P2 cartesian 2d mesh", build_parabolic_mesh(MeshDataBaseForTests::get().cartesian2DMesh())},
                 NamedMesh{"hybrid 2d mesh", MeshDataBaseForTests::get().hybrid2DMesh()},
                 NamedMesh{"P2 hybrid 2d mesh", build_parabolic_mesh(MeshDataBaseForTests::get().hybrid2DMesh())}};

    auto descriptor = PolynomialReconstructionDescriptor{IntegrationMethodType::boundary, 2};

    SECTION(name(descriptor.integrationMethodType()))
    {
      SECTION("2D")
      {
        std::string_view data = R"(
import math;
let scalar_2d: R^2 -> R, x -> 2*x[0]*x[0]-x[0]*x[1]+1.5*x[1]*x[1] + 3*x[0]-2*x[1] + 2;
let R3_2d: R^2 -> R^3, x -> [2*exp(x[0])*sin(x[1])+3, x[0]-2*x[1], 3];
let R2x2_2d: R^2 -> R^2x2, x -> [[2*exp(x[0])*sin(x[1])+3, sin(x[0]-2*x[1])], [3, x[0]*x[1]]];
)";

        TAO_PEGTL_NAMESPACE::string_input input{data, "test.pgs"};

        auto ast = ASTBuilder::build(input);

        ASTModulesImporter{*ast};
        ASTNodeTypeCleaner<language::import_instruction>{*ast};

        ASTSymbolTableBuilder{*ast};
        ASTNodeDataTypeBuilder{*ast};

        ASTNodeTypeCleaner<language::var_declaration>{*ast};
        ASTNodeTypeCleaner<language::fct_declaration>{*ast};
        ASTNodeExpressionBuilder{*ast};

        std::shared_ptr<SymbolTable> symbol_table = ast->m_symbol_table;

        // ensure that variables are declared at this point
        TAO_PEGTL_NAMESPACE::position position{data.size(), 1, 1, "fixture"};

        auto [i_symbol, found] = symbol_table->find("scalar_2d", position);
        REQUIRE(found);
        REQUIRE(i_symbol->attributes().dataType() == ASTNodeDataType::function_t);

        FunctionSymbolId function_symbol_id(std::get<uint64_t>(i_symbol->attributes().value()), symbol_table);

        SECTION("R data")
        {
          for (auto named_mesh : mesh_list) {
            SECTION(named_mesh.name())
            {
              // polynomial is degree 2, transformation is degree 2 and detJ is
              // degree 3 => 7
              auto quadrature_descriptor = GaussLegendreQuadratureDescriptor(7);

              auto mesh_v = named_mesh.mesh();
              std::visit(
                [=](auto&& p_mesh) {
                  using MeshType = mesh_type_t<decltype(p_mesh)>;
                  if constexpr (MeshType::Dimension == 2) {
                    using Rd = TinyVector<MeshType::Dimension>;

                    auto p = [](const Rd& x) {
                      return 2 * x[0] * x[0] - x[0] * x[1] + 1.5 * x[1] * x[1] + 3 * x[0] - 2 * x[1] + 2;
                    };

                    ASTNode::setStackDetails(false);

                    DiscreteFunctionP0<double> fh(p_mesh);
                    IntegrateOnCells<double(Rd)>::integrateTo(function_symbol_id, quadrature_descriptor, *p_mesh, fh);

                    ASTNode::setStackDetails(true);

                    auto Vj = MeshDataManager::instance().getMeshData(*p_mesh).Vj();

                    parallel_for(
                      p_mesh->numberOfCells(), PUGS_LAMBDA(const CellId& cell_id) { fh[cell_id] /= Vj[cell_id]; });

                    auto reconstructions = PolynomialReconstruction{descriptor}.build(fh);

                    auto dpk_fh = reconstructions[0]->get<DiscreteFunctionDPk<MeshType::Dimension, const double>>();

                    double error_sum = 0;

                    auto cell_to_node_matrix = p_mesh->connectivity().cellToNodeMatrix();
                    auto xr                  = p_mesh->xr();

                    for (CellId cell_id = 0; cell_id < p_mesh->numberOfCells(); ++cell_id) {
                      const Rd& x0 = xr[cell_to_node_matrix[cell_id][0]];
                      const Rd& x1 = xr[cell_to_node_matrix[cell_id][1]];
                      const Rd& x2 = xr[cell_to_node_matrix[cell_id][2]];

                      for (size_t i = 0; i < 5; ++i) {
                        const double theta = i / 4.;
                        for (size_t j = 0; j < 5; ++j) {
                          const double psi = j / 4.;

                          const Rd X = theta * x0 + (1 - theta) * (psi * x1 + (1 - psi) * x2);

                          error_sum += Vj[cell_id] * std::abs(p(X) - dpk_fh[cell_id](X));
                        }
                      }
                    }

                    REQUIRE(error_sum == Catch::Approx(0.).margin(1E-11));
                  }
                },
                mesh_v->variant());
            }
          }
        }
      }
    }
  }
}
