#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/GaussLobattoQuadratureDescriptor.hpp>
#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/SquareQ2Transformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("SquareQ2Transformation", "[geometry]")
{
  SECTION("2D")
  {
    using R2 = TinyVector<2>;

    const R2 a00_hat{-1, -1};
    const R2 a01_hat{-1, +0};
    const R2 a02_hat{-1, +1};

    const R2 a10_hat{+0, -1};
    const R2 a11_hat{+0, +0};
    const R2 a12_hat{+0, +1};

    const R2 a20_hat{+1, -1};
    const R2 a21_hat{+1, +0};
    const R2 a22_hat{+1, +1};

    const R2 a00{0, 0};
    const R2 a01{0.2, 1};
    const R2 a02{0.3, 2.5};

    const R2 a10{0.7, -0.2};
    const R2 a11{0.9, 0.8};
    const R2 a12{1.1, 2.1};

    const R2 a20{1.7, 0.1};
    const R2 a21{1.9, 1.2};
    const R2 a22{2.1, 2.1};

    const SquareQ2Transformation<2> t{a00, a01, a02,   //
                                      a10, a11, a12,   //
                                      a20, a21, a22};

    SECTION("values")
    {
      REQUIRE(t(a00_hat)[0] == Catch::Approx(a00[0]));
      REQUIRE(t(a00_hat)[1] == Catch::Approx(a00[1]));

      REQUIRE(t(a01_hat)[0] == Catch::Approx(a01[0]));
      REQUIRE(t(a01_hat)[1] == Catch::Approx(a01[1]));

      REQUIRE(t(a02_hat)[0] == Catch::Approx(a02[0]));
      REQUIRE(t(a02_hat)[1] == Catch::Approx(a02[1]));

      REQUIRE(t(a10_hat)[0] == Catch::Approx(a10[0]));
      REQUIRE(t(a10_hat)[1] == Catch::Approx(a10[1]));

      REQUIRE(t(a11_hat)[0] == Catch::Approx(a11[0]));
      REQUIRE(t(a11_hat)[1] == Catch::Approx(a11[1]));

      REQUIRE(t(a12_hat)[0] == Catch::Approx(a12[0]));
      REQUIRE(t(a12_hat)[1] == Catch::Approx(a12[1]));

      REQUIRE(t(a20_hat)[0] == Catch::Approx(a20[0]));
      REQUIRE(t(a20_hat)[1] == Catch::Approx(a20[1]));

      REQUIRE(t(a21_hat)[0] == Catch::Approx(a21[0]));
      REQUIRE(t(a21_hat)[1] == Catch::Approx(a21[1]));

      REQUIRE(t(a22_hat)[0] == Catch::Approx(a22[0]));
      REQUIRE(t(a22_hat)[1] == Catch::Approx(a22[1]));
    }

    SECTION("Jacobian determinant")
    {
      // Exact calculations in this section have been obtain with Maxima
      SECTION("at points")
      {
        auto detJ = [](const R2& X) {
          const double x  = X[0];
          const double y  = X[1];
          const double x2 = x * x;
          const double y2 = y * y;
          const double x3 = x2 * x;
          const double y3 = y2 * y;

          return 0.0125 * x2 * y3 - 0.015 * x * y3 + 0.0075 * y3 + 0.0025 * x3 * y2 + 0.015 * x2 * y2 -
                 0.04375 * x * y2 + 0.07125 * y2 - 0.015 * x3 * y - 0.25375 * x2 * y - 0.26 * x * y + 0.30875 * y +
                 0.0075 * x3 - 0.07125 * x2 + 0.11625 * x + 0.9575;
        };

        REQUIRE(t.jacobianDeterminant(a00_hat) == Catch::Approx(detJ(a00_hat)));
        REQUIRE(t.jacobianDeterminant(a01_hat) == Catch::Approx(detJ(a01_hat)));
        REQUIRE(t.jacobianDeterminant(a02_hat) == Catch::Approx(detJ(a02_hat)));

        REQUIRE(t.jacobianDeterminant(a10_hat) == Catch::Approx(detJ(a10_hat)));
        REQUIRE(t.jacobianDeterminant(a11_hat) == Catch::Approx(detJ(a11_hat)));
        REQUIRE(t.jacobianDeterminant(a12_hat) == Catch::Approx(detJ(a12_hat)));

        REQUIRE(t.jacobianDeterminant(a20_hat) == Catch::Approx(detJ(a20_hat)));
        REQUIRE(t.jacobianDeterminant(a21_hat) == Catch::Approx(detJ(a21_hat)));
        REQUIRE(t.jacobianDeterminant(a22_hat) == Catch::Approx(detJ(a22_hat)));

        const R2 X0_hat{-0.7, 0.4};
        REQUIRE(t.jacobianDeterminant(X0_hat) == Catch::Approx(detJ(X0_hat)));

        const R2 X1_hat{0.23, -0.12};
        REQUIRE(t.jacobianDeterminant(X1_hat) == Catch::Approx(detJ(X1_hat)));

        const R2 X2_hat{0.32, 0.91};
        REQUIRE(t.jacobianDeterminant(X2_hat) == Catch::Approx(detJ(X2_hat)));
      }

      SECTION("Gauss order 3")
      {
        // Jacobian determinant is a polynomial function of degree 3
        const QuadratureFormula<2>& gauss =
          QuadratureManager::instance().getSquareFormula(GaussLegendreQuadratureDescriptor(3));

        double surface = 0;
        for (size_t i = 0; i < gauss.numberOfPoints(); ++i) {
          surface += gauss.weight(i) * t.jacobianDeterminant(gauss.point(i));
        }

        // 1151. / 300 is actually the surface of the Q2-quadrangle
        REQUIRE(surface == Catch::Approx(1151. / 300));
      }
    }
  }
}
