#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussLegendreQuadratureDescriptor.hpp>
#include <analysis/GaussLobattoQuadratureDescriptor.hpp>
#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/SquareQ3Transformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("SquareQ3Transformation", "[geometry]")
{
  SECTION("2D")
  {
    using R2 = TinyVector<2>;

    const R2 a00_hat{-1, -1};
    const R2 a01_hat{-1, -1. / 3};
    const R2 a02_hat{-1, 1. / 3};
    const R2 a03_hat{-1, +1};

    const R2 a10_hat{-1. / 3, -1};
    const R2 a11_hat{-1. / 3, -1. / 3};
    const R2 a12_hat{-1. / 3, +1. / 3};
    const R2 a13_hat{-1. / 3, +1};

    const R2 a20_hat{+1. / 3, -1};
    const R2 a21_hat{+1. / 3, -1. / 3};
    const R2 a22_hat{+1. / 3, +1. / 3};
    const R2 a23_hat{+1. / 3, +1};

    const R2 a30_hat{+1, -1.};
    const R2 a31_hat{+1, -1. / 3};
    const R2 a32_hat{+1, +1. / 3};
    const R2 a33_hat{+1, +1};

    const R2 a00{-1, -1};
    const R2 a01{-1.2, -0.55};
    const R2 a02{-1, 0.18};
    const R2 a03{-0.7, 1};

    const R2 a10{-0.12, -1.1};
    const R2 a11{-0.18, -0.54};
    const R2 a12{-0.28, 0.1};
    const R2 a13{-0.36, 1.2};

    const R2 a20{0.24, -0.7};
    const R2 a21{0.24, -0.1};
    const R2 a22{0.4, 0.51};
    const R2 a23{0.36, 0.96};

    const R2 a30{0.84, -0.94};
    const R2 a31{0.81, -0.5};
    const R2 a32{1.02, 0.22};
    const R2 a33{1.3, 0.9};

    const SquareQ3Transformation<2> t{a00, a01, a02, a03,   //
                                      a10, a11, a12, a13,   //
                                      a20, a21, a22, a23,   //
                                      a30, a31, a32, a33};

    SECTION("values")
    {
      REQUIRE(t(a00_hat)[0] == Catch::Approx(a00[0]));
      REQUIRE(t(a00_hat)[1] == Catch::Approx(a00[1]));

      REQUIRE(t(a01_hat)[0] == Catch::Approx(a01[0]));
      REQUIRE(t(a01_hat)[1] == Catch::Approx(a01[1]));

      REQUIRE(t(a02_hat)[0] == Catch::Approx(a02[0]));
      REQUIRE(t(a02_hat)[1] == Catch::Approx(a02[1]));

      REQUIRE(t(a03_hat)[0] == Catch::Approx(a03[0]));
      REQUIRE(t(a03_hat)[1] == Catch::Approx(a03[1]));

      REQUIRE(t(a10_hat)[0] == Catch::Approx(a10[0]));
      REQUIRE(t(a10_hat)[1] == Catch::Approx(a10[1]));

      REQUIRE(t(a11_hat)[0] == Catch::Approx(a11[0]));
      REQUIRE(t(a11_hat)[1] == Catch::Approx(a11[1]));

      REQUIRE(t(a12_hat)[0] == Catch::Approx(a12[0]));
      REQUIRE(t(a12_hat)[1] == Catch::Approx(a12[1]));

      REQUIRE(t(a13_hat)[0] == Catch::Approx(a13[0]));
      REQUIRE(t(a13_hat)[1] == Catch::Approx(a13[1]));

      REQUIRE(t(a20_hat)[0] == Catch::Approx(a20[0]));
      REQUIRE(t(a20_hat)[1] == Catch::Approx(a20[1]));

      REQUIRE(t(a21_hat)[0] == Catch::Approx(a21[0]));
      REQUIRE(t(a21_hat)[1] == Catch::Approx(a21[1]));

      REQUIRE(t(a22_hat)[0] == Catch::Approx(a22[0]));
      REQUIRE(t(a22_hat)[1] == Catch::Approx(a22[1]));

      REQUIRE(t(a23_hat)[0] == Catch::Approx(a23[0]));
      REQUIRE(t(a23_hat)[1] == Catch::Approx(a23[1]));

      REQUIRE(t(a30_hat)[0] == Catch::Approx(a30[0]));
      REQUIRE(t(a30_hat)[1] == Catch::Approx(a30[1]));

      REQUIRE(t(a31_hat)[0] == Catch::Approx(a31[0]));
      REQUIRE(t(a31_hat)[1] == Catch::Approx(a31[1]));

      REQUIRE(t(a32_hat)[0] == Catch::Approx(a32[0]));
      REQUIRE(t(a32_hat)[1] == Catch::Approx(a32[1]));

      REQUIRE(t(a33_hat)[0] == Catch::Approx(a33[0]));
      REQUIRE(t(a33_hat)[1] == Catch::Approx(a33[1]));
    };

    SECTION("Jacobian determinant")
    {
      // Exact calculations in this section have been obtain with Maxima
      SECTION("at points")
      {
        auto detJ = [](const R2& X) {
          const double x  = X[0];
          const double y  = X[1];
          const double x2 = x * x;
          const double y2 = y * y;
          const double x3 = x2 * x;
          const double y3 = y2 * y;
          const double x4 = x3 * x;
          const double y4 = y3 * y;
          const double x5 = x4 * x;
          const double y5 = y4 * y;

          return -0.222971484374999 * x4 * y5 - 0.2019878173828124 * x3 * y5 + 0.4812361083984359 * x2 * y5 -
                 0.1343915771484368 * x * y5 - 0.2237842529296873 * y5 - 0.8692403961181647 * x5 * y4 +
                 0.2119390411376944 * x4 * y4 + 0.92333919067383 * x3 * y4 + 0.3825069763183597 * x2 * y4 -
                 0.1489811187744143 * x * y4 - 0.2874291229248046 * y4 - 2.052875390625002 * x5 * y3 +
                 0.95425850830078 * x4 * y3 + 2.841806689453126 * x3 * y3 - 0.106753491210934 * x2 * y3 -
                 0.459694775390626 * x * y3 - 0.2465528686523447 * y3 + 2.632395684814454 * x5 * y2 -
                 0.8694205993652319 * x4 * y2 - 1.642588439941411 * x3 * y2 - 1.819210363769531 * x2 * y2 -
                 0.2746851745605458 * x * y2 + 0.532296002197265 * y2 + 0.5854069335937499 * x5 * y +
                 0.7183677612304681 * x4 * y + 0.3296906982421864 * x3 * y - 1.984359375000001 * x2 * y -
                 0.5961610107421865 * x * y + 0.8772590942382814 * y - 1.584013980102539 * x5 +
                 0.5495632965087884 * x4 + 2.188325225830079 * x3 + 0.8013367858886723 * x2 - 0.895632241821289 * x +
                 0.714292984008789;
        };

        REQUIRE(t.jacobianDeterminant(a00_hat) == Catch::Approx(detJ(a00_hat)));
        REQUIRE(t.jacobianDeterminant(a01_hat) == Catch::Approx(detJ(a01_hat)));
        REQUIRE(t.jacobianDeterminant(a02_hat) == Catch::Approx(detJ(a02_hat)));

        REQUIRE(t.jacobianDeterminant(a10_hat) == Catch::Approx(detJ(a10_hat)));
        REQUIRE(t.jacobianDeterminant(a11_hat) == Catch::Approx(detJ(a11_hat)));
        REQUIRE(t.jacobianDeterminant(a12_hat) == Catch::Approx(detJ(a12_hat)));

        REQUIRE(t.jacobianDeterminant(a20_hat) == Catch::Approx(detJ(a20_hat)));
        REQUIRE(t.jacobianDeterminant(a21_hat) == Catch::Approx(detJ(a21_hat)));
        REQUIRE(t.jacobianDeterminant(a22_hat) == Catch::Approx(detJ(a22_hat)));

        const R2 X0_hat{-0.7, 0.4};
        REQUIRE(t.jacobianDeterminant(X0_hat) == Catch::Approx(detJ(X0_hat)));

        const R2 X1_hat{0.23, -0.12};
        REQUIRE(t.jacobianDeterminant(X1_hat) == Catch::Approx(detJ(X1_hat)));

        const R2 X2_hat{0.32, 0.91};
        REQUIRE(t.jacobianDeterminant(X2_hat) == Catch::Approx(detJ(X2_hat)));
      }

      SECTION("Gauss order 5")
      {
        // Jacobian determinant is a polynomial function of degree 5 (= ordre des polynomes derives)
        const QuadratureFormula<2>& gauss =
          QuadratureManager::instance().getSquareFormula(GaussLegendreQuadratureDescriptor(5));

        double surface = 0;
        for (size_t i = 0; i < gauss.numberOfPoints(); ++i) {
          surface += gauss.weight(i) * t.jacobianDeterminant(gauss.point(i));
        }

        // the surface of the Q3-quadrangle
        REQUIRE(surface == Catch::Approx(2909504171292913893091170139541425618159185098. /
                                         738343208252814099208113646063313420356936035.));
      }
    }
  }
}
