#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/TriangleP2Transformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("TriangleP2Transformation", "[geometry]")
{
  SECTION("2D")
  {
    using R2 = TinyVector<2>;

    const R2 a0{1, 2};
    const R2 a1{3, 1};
    const R2 a2{2, 5};

    const R2 a01{2, 1.3};
    const R2 a12{2.3, 4};
    const R2 a20{1.6, 3.6};

    const TriangleP2Transformation<2> t(a0, a1, a2,   //
                                        a12, a20, a01);

    REQUIRE(t(R2{0, 0})[0] == Catch::Approx(a0[0]));
    REQUIRE(t(R2{0, 0})[1] == Catch::Approx(a0[1]));

    REQUIRE(t(R2{1, 0})[0] == Catch::Approx(a1[0]));
    REQUIRE(t(R2{1, 0})[1] == Catch::Approx(a1[1]));

    REQUIRE(t(R2{0, 1})[0] == Catch::Approx(a2[0]));
    REQUIRE(t(R2{0, 1})[1] == Catch::Approx(a2[1]));

    REQUIRE(t(R2{0.5, 0})[0] == Catch::Approx(a01[0]));
    REQUIRE(t(R2{0.5, 0})[1] == Catch::Approx(a01[1]));

    REQUIRE(t(R2{0.5, 0.5})[0] == Catch::Approx(a12[0]));
    REQUIRE(t(R2{0.5, 0.5})[1] == Catch::Approx(a12[1]));

    REQUIRE(t(R2{0, 0.5})[0] == Catch::Approx(a20[0]));
    REQUIRE(t(R2{0, 0.5})[1] == Catch::Approx(a20[1]));

    // Values obtained using Maxima
    const R2 X0_hat{1. / 3, 1. / 3};
    REQUIRE(t(X0_hat)[0] == Catch::Approx(88. / 45));
    REQUIRE(t(X0_hat)[1] == Catch::Approx(46. / 15));

    const R2 X1_hat{0.7, 0.2};
    REQUIRE(t(X1_hat)[0] == Catch::Approx(312. / 125));
    REQUIRE(t(X1_hat)[1] == Catch::Approx(603. / 250));

    const R2 X2_hat{0.3, 0.6};
    REQUIRE(t(X2_hat)[0] == Catch::Approx(52. / 25));
    REQUIRE(t(X2_hat)[1] == Catch::Approx(211. / 50));

    auto detJ = [](const R2& X) {
      const double x  = X[0];
      const double y  = X[1];
      const double x2 = x * x;
      const double xy = x * y;
      const double y2 = y * y;

      return 4.48 * y2 + 1.28 * xy - 13.28 * y + 1.92 * x2 + 4.4 * x + 9.32;
    };

    REQUIRE(t.jacobianDeterminant(a0) == Catch::Approx(detJ(a0)));
    REQUIRE(t.jacobianDeterminant(a1) == Catch::Approx(detJ(a1)));
    REQUIRE(t.jacobianDeterminant(a2) == Catch::Approx(detJ(a2)));

    REQUIRE(t.jacobianDeterminant(a01) == Catch::Approx(detJ(a01)));
    REQUIRE(t.jacobianDeterminant(a12) == Catch::Approx(detJ(a12)));
    REQUIRE(t.jacobianDeterminant(a20) == Catch::Approx(detJ(a20)));

    REQUIRE(t.jacobianDeterminant(X0_hat) == Catch::Approx(detJ(X0_hat)));
    REQUIRE(t.jacobianDeterminant(X1_hat) == Catch::Approx(detJ(X1_hat)));
    REQUIRE(t.jacobianDeterminant(X2_hat) == Catch::Approx(detJ(X2_hat)));

    QuadratureFormula<2> qf = QuadratureManager::instance().getTriangleFormula(GaussQuadratureDescriptor(2));

    double surface = 0;
    for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
      surface += qf.weight(i) * t.jacobianDeterminant(qf.point(i));
    }

    // 113. / 30 is actually the surface of the P2-quadrangle
    // (computed with Maxima)
    REQUIRE(surface == Catch::Approx(113. / 30));
  }
}
