#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>

#include <analysis/GaussQuadratureDescriptor.hpp>
#include <analysis/QuadratureManager.hpp>
#include <geometry/TriangleP3Transformation.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("TriangleP3Transformation", "[geometry]")
{
  SECTION("2D")
  {
    using R2 = TinyVector<2>;

    const R2 a0{1, 2};
    const R2 a1{5, 1};
    const R2 a2{2, 4};

    const R2 a001{4. / 3, 1.6};
    const R2 a011{8. / 3, 0.8};

    const R2 a112{3.9, 1.9};
    const R2 a122{3.2, 2.9};

    const R2 a220{1.8, 3.2};
    const R2 a200{3.2, 2.9};

    const R2 a012{2.5, 2};

    const TriangleP3Transformation<2> t(a0, a1, a2,   //
                                        a112, a122,   //
                                        a220, a200,   //
                                        a001, a011,   //
                                        a012);

    const R2 a0_hat{0, 0};
    const R2 a1_hat{1, 0};
    const R2 a2_hat{0, 1};

    const R2 a001_hat{1. / 3, 0};
    const R2 a011_hat{2. / 3, 0};

    const R2 a112_hat{2. / 3, 1. / 3};
    const R2 a122_hat{1. / 3, 2. / 3};

    const R2 a220_hat{0, 2. / 3};
    const R2 a200_hat{0, 1. / 3};

    const R2 a012_hat{1. / 3, 1. / 3};

    CHECK(t(a0_hat)[0] == Catch::Approx(a0[0]));
    CHECK(t(a0_hat)[1] == Catch::Approx(a0[1]));

    CHECK(t(a1_hat)[0] == Catch::Approx(a1[0]));
    CHECK(t(a1_hat)[1] == Catch::Approx(a1[1]));

    CHECK(t(a2_hat)[0] == Catch::Approx(a2[0]));
    CHECK(t(a2_hat)[1] == Catch::Approx(a2[1]));

    CHECK(t(a001_hat)[0] == Catch::Approx(a001[0]));
    CHECK(t(a001_hat)[1] == Catch::Approx(a001[1]));

    CHECK(t(a011_hat)[0] == Catch::Approx(a011[0]));
    CHECK(t(a011_hat)[1] == Catch::Approx(a011[1]));

    CHECK(t(a122_hat)[0] == Catch::Approx(a122[0]));
    CHECK(t(a122_hat)[1] == Catch::Approx(a122[1]));

    CHECK(t(a112_hat)[0] == Catch::Approx(a112[0]));
    CHECK(t(a112_hat)[1] == Catch::Approx(a112[1]));
    //
    CHECK(t(a220_hat)[0] == Catch::Approx(a220[0]));
    CHECK(t(a220_hat)[1] == Catch::Approx(a220[1]));

    CHECK(t(a200_hat)[0] == Catch::Approx(a200[0]));
    CHECK(t(a200_hat)[1] == Catch::Approx(a200[1]));

    CHECK(t(a012_hat)[0] == Catch::Approx(a012[0]));
    CHECK(t(a012_hat)[1] == Catch::Approx(a012[1]));

    // Values obtained using Maxima

    const R2 X0_hat{0.2, 0.3};
    REQUIRE(t(X0_hat)[0] == Catch::Approx(2.5464));
    REQUIRE(t(X0_hat)[1] == Catch::Approx(2.31225));

    const R2 X1_hat{0.6, 0.1};
    REQUIRE(t(X1_hat)[0] == Catch::Approx(2.7548));
    REQUIRE(t(X1_hat)[1] == Catch::Approx(1.10455));

    const R2 X2_hat{0.8, 0.05};
    REQUIRE(t(X2_hat)[0] == Catch::Approx(3.669725));
    REQUIRE(t(X2_hat)[1] == Catch::Approx(0.8521937500000003));

    auto detJ = [](const R2& X) {
      const double x  = X[0];
      const double y  = X[1];
      const double x2 = x * x;
      const double y2 = y * y;
      const double x3 = x * x * x;
      const double x4 = x * x * x * x;
      const double y3 = y * y * y;
      const double y4 = y * y * y * y;

      return -414.315 * y4 - 1833.435 * x * y3 + 1150.4025 * y3 - 2720.9925 * x2 * y2 + 3589.7175 * x * y2 -
             862.56 * y2 - 1598.94 * x3 * y + 3373.4475 * x2 * y - 1920.96 * x * y + 193.185 * y - 280.665 * x4 +
             922.185 * x3 - 937.98 * x2 + 351.045 * x - 16.11;
    };

    REQUIRE(t.jacobianDeterminant(X0_hat) == Catch::Approx(detJ(X0_hat)));
    REQUIRE(t.jacobianDeterminant(X1_hat) == Catch::Approx(detJ(X1_hat)));
    REQUIRE(t.jacobianDeterminant(X2_hat) == Catch::Approx(detJ(X2_hat)));

    QuadratureFormula<2> qf = QuadratureManager::instance().getTriangleFormula(GaussQuadratureDescriptor(4));

    double surface = 0;
    for (size_t i = 0; i < qf.numberOfPoints(); ++i) {
      surface += qf.weight(i) * t.jacobianDeterminant(qf.point(i));
    }

    // 113. / 30 is actually the surface of the P3-quadrangle
    // (computed with Maxima)
    REQUIRE(surface == Catch::Approx(42891. / 8000));
  }
}
